#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Focus Telescope Recipe for Megara"""

from __future__ import division, print_function

import numpy
from numina.array import combine
from numina.core.dataholders import Result, Parameter, Requirement
from numina.core.requirements import ObservationResultRequirement
from numina.exceptions import RecipeError

from megaradrp.recipes.scientific.base import ImageRecipe
import megaradrp.requirements as reqs
from megaradrp.processing.combine import basic_processing_with_combination_frames


class FocusTelescopeRecipe(ImageRecipe):
    """Process telescope focus images and find best focus.

    This recipe process a set of focus images obtained in
    **Focus Telescope** mode and returns an estimation
    of the telescope best focus.

    See Also
    --------
    megaradrp.recipes.auxiliary.focusspec.FocusSpectrographRecipe:
                recipe to measure the focus of the spectrograph

    Notes
    -----
    Images provided in `obresult` are grouped by the value of their
    FOCUST keyword. Groups of images are trimmed and corrected from overscan,
    bad pixel mask (if `master_bpm` is not None), bias and dark current
    (if `master_dark` is not None). Each group is then stacked using the median.

    The result of the combination is saved as an intermediate result, named
    'focus2d-#focus.fits', with #focus being the value of the focus
    of each group. Apertures are extracted in each combined image, and the
    resulting RSS file is saved as an intermediate result, named
    'focus1-#focus.fits'.

    For each image, the FWHM of the object at `position` is computed.

    Then, the FWHM  is fitted to a 2nd degree polynomial, and the focus
    corresponding to its minimum is obtained and returned in `focus_table`

    """

    # Requirements
    obresult = ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_bpm = reqs.MasterBPMRequirement()
    master_traces = reqs.MasterAperturesRequirement()
    extraction_offset = Parameter([0.0], 'Offset traces for extraction', accept_scalar=True)
    master_wlcalib = reqs.WavelengthCalibrationRequirement()
    position = Requirement(list, "Position of the reference object", default=(0, 0))
    # Products
    focus_table = Result(float)

    # @numina.core.validator.validate
    def run(self, rinput):
        # Basic processing
        self.logger.info('start focus telescope')

        obresult = rinput.obresult

        flow = self.init_filters(rinput, obresult.configuration)

        coors = rinput.position
        focus_t = 'M2UZ'

        image_groups = {}
        self.logger.info('group images by focus')

        for idx, frame in enumerate(obresult.frames):
            with frame.open() as img:
                focus_val = img[0].header[focus_t]
                if focus_val not in image_groups:
                    self.logger.debug('new focus %s', focus_val)
                    image_groups[focus_val] = []
                self.logger.debug('image %s in group %s', img, focus_val)
                image_groups[focus_val].append(frame)

        if len(image_groups) < 2:
            raise RecipeError('We have only {} different focus'.format(len(image_groups)))

        all_images = {}
        for focus, frames in image_groups.items():
            self.logger.info('processing focus %s', focus)

            try:
                img = basic_processing_with_combination_frames(frames, flow, method=combine.median, errors=False)

                self.save_intermediate_img(img, 'focus2d-%s.fits' % (focus,))

                # 1D, extraction, Wl calibration, Flat fielding
                _, img1d = self.run_reduction_1d(
                    img,
                    rinput.master_traces,
                    rinput.master_wlcalib,
                    rinput.master_fiberflat,
                    offset=rinput.extraction_offset
                )

                do_sky_subtraction = True
                if do_sky_subtraction:
                    self.logger.info('start sky subtraction')
                    final, origin, sky = self.run_sky_subtraction(
                        img1d, rinput.ignored_sky_bundles)
                    self.logger.info('end sky subtraction')
                else:
                    final = img1d
                    origin = final
                    sky = final

                self.save_intermediate_img(final, 'focus1d-%s.fits' % (focus,))

                self.logger.info('find lines and compute FWHM')
                star_rss_fwhm = self.run_on_image(final, coors)
                all_images[focus] = star_rss_fwhm

            except ValueError:
                self.logger.info('focus %s cannot be processed', focus)

        self.logger.info('fit FWHM of star')
        final = self.reorder_and_fit(all_images)
        self.logger.info('best focus is %s', final)

        self.logger.info('end focus telescope')
        return self.create_result(focus_table=final)

    def run_on_image(self, img, coors):
        """Extract spectra, find peaks and compute FWHM."""

        from scipy.spatial import KDTree

        fiberconf = self.datamodel.get_fiberconf(img)
        self.logger.debug("LCB configuration is %s", fiberconf.conf_id)
        rssdata = img[0].data
        cut1 = 1000
        cut2 = 3000
        points = [(0, 0)] # Center of fiber 313
        fibers = fiberconf.conected_fibers(valid_only=True)
        grid_coords = []
        for fiber in fibers:
            grid_coords.append((fiber.x, fiber.y))
        # setup kdtree for searching
        kdtree = KDTree(grid_coords)

        # Other posibility is
        # query using radius instead
        # radius = 1.2
        # kdtree.query_ball_point(points, k=7, r=radius)

        npoints = 19 + 18
        # 1 + 6  for first ring
        # 1 + 6  + 12  for second ring
        # 1 + 6  + 12  + 18 for third ring
        dis_p, idx_p = kdtree.query(points, k=npoints)

        self.logger.info('Using %d nearest fibers', npoints)
        for diss, idxs, point in zip(dis_p, idx_p, points):
            # For each point
            self.logger.info('For point %s', point)
            colids = []
            coords = []
            for dis, idx in zip(diss, idxs):
                fiber = fibers[idx]
                colids.append(fiber.fibid - 1)
                coords.append((fiber.x, fiber.y))

            coords = numpy.asarray(coords)
            flux_per_cell = rssdata[colids, cut1:cut2].mean(axis=1)
            flux_per_cell_total = flux_per_cell.sum()
            flux_per_cell_norm = flux_per_cell / flux_per_cell_total
            # centroid
            scf = coords.T * flux_per_cell_norm
            centroid = scf.sum(axis=1)
            self.logger.info('centroid: %s', centroid)
            # central coords
            c_coords = coords - centroid
            scf0 = scf - centroid[:, numpy.newaxis] * flux_per_cell_norm
            mc2 = numpy.dot(scf0, c_coords)
            self.logger.info('2nd order moments, x2=%f, y2=%f, xy=%f', mc2[0,0], mc2[1,1], mc2[0,1])

        # FIXME: returning only 1 value for 1 star
        return mc2[0,0]

    def reorder_and_fit(self, all_images):
        """Fit all the values of FWHM to a 2nd degree polynomial and return minimum."""

        # We are assuming there is only 1 star
        focii = sorted(all_images.keys())

        ally = [all_images[focus] for focus in focii]

        try:
            res = numpy.polyfit(focii, ally, deg=2)
            self.logger.debug('fitting to deg 2 polynomial, done')
            self.logger.debug('parameters are %s', res)
            best = -res[1] / (2 * res[0])
        except ValueError as error:
            self.logger.warning("Error in fitting: %s", error)
            best = 0.0

        return best




# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Extract objects from RSS images"""

import math
import uuid
import logging

import numpy
import astropy.wcs
import astropy.io.fits as fits
import astropy.units as u
from scipy.spatial import KDTree
from scipy.ndimage.filters import gaussian_filter
from numina.frame.utils import copy_img

from megaradrp.processing.fluxcalib import update_flux_limits


def coverage_det(arr):
    """Compute coverage"""

    ma = numpy.max(arr)
    mn = numpy.min(arr)

    maxcover = len(arr) + 1

    if ma == 0:
        return slice(0, 0)

    if mn == 1:
        return slice(0, maxcover)

    # We have 0s and 1s
    diffa = numpy.diff(arr)

    # For the left, find first +1
    val_f, = numpy.where(diffa==1)
    pix_f = 0
    if len(val_f) > 0:
        pix_f = val_f[0] + 1

    # For the rigth, find first -1
    val_l, = numpy.where(diffa==-1)
    pix_l = maxcover
    if len(val_l) > 0:
        pix_l = val_l[0] + 1

    return slice(pix_f, pix_l)


def extract_star(rssimage, position, npoints, fiberconf, logger=None):
    """Extract a star given its center and the number of fibers to extract"""

    # FIXME: handle several positions

    logger.info('extracting star')

    # fiberconf = datamodel.get_fiberconf(rssimage)
    logger.debug("Configuration UUID is %s", fiberconf.conf_id)

    rssdata = rssimage[0].data
    pdata = rssimage['wlmap'].data

    points = [position]
    fibers = fiberconf.conected_fibers(valid_only=True)
    grid_coords = []
    for fiber in fibers:
        grid_coords.append((fiber.x, fiber.y))
    # setup kdtree for searching
    kdtree = KDTree(grid_coords)

    # Other posibility is
    # query using radius instead
    # radius = 1.2
    # kdtree.query_ball_point(points, k=7, r=radius)

    dis_p, idx_p = kdtree.query(points, k=npoints)

    logger.info('Using %d nearest fibers', npoints)
    totals = []
    for diss, idxs, point in zip(dis_p, idx_p, points):
        # For each point
        logger.info('For point %s', point)
        colids = []
        coords = []
        for dis, idx in zip(diss, idxs):
            fiber = fibers[idx]
            colids.append(fiber.fibid - 1)
            logger.debug('adding fibid %d', fiber.fibid)

            coords.append((fiber.x, fiber.y))

        colids.sort()
        flux_fiber = rssdata[colids]
        flux_total = rssdata[colids].sum(axis=0)
        coverage_total = pdata[colids].sum(axis=0)

        max_cover = coverage_total.max()
        some_value_region = coverage_total > 0
        max_value_region = coverage_total == max_cover
        valid_region = max_value_region

        # Interval with maximum coverage
        #nz_max, = numpy.nonzero(numpy.diff(max_value_region))
        # Interval with at least 1 fiber
        #nz_some, = numpy.nonzero(numpy.diff(some_value_region))
        nz_max_slice = coverage_det(max_value_region.astype('int'))
        nz_some_slice = coverage_det(some_value_region.astype('int'))

        # Collapse the flux in the optimal region
        perf = flux_fiber[:, nz_max_slice].sum(axis=1)
        # Contribution of each fiber to the total flux, 1D
        perf_norm = perf / perf.sum()
        contributions = numpy.zeros(shape=(rssdata.shape[0],))
        contributions[colids] = perf_norm
        # Contribution of each fiber to the total flux, 2D
        flux_per_fiber = pdata * contributions[:, numpy.newaxis]
        flux_sum = flux_per_fiber.sum(axis=0)
        # In the region max_value_region, flux_sum == 1
        # In some_value_region 0 < flux_sum < 1
        # Outside is flux_sum == 0
        flux_correction = numpy.zeros_like(flux_sum)
        flux_correction[some_value_region] = 1.0 / flux_sum[some_value_region]

        # Limit to 10
        flux_correction = numpy.clip(flux_correction, 0, 10)

        flux_total_c = flux_total * flux_correction
        # plt.axvspan(nz_some[0], nz_some[1], alpha=0.2, facecolor='r')
        # plt.axvspan(nz_max[0], nz_max[1], alpha=0.2, facecolor='r')
        # plt.plot(flux_correction)
        # plt.show()
        #
        # plt.axvspan(nz_some[0], nz_some[1], alpha=0.2)
        # plt.axvspan(nz_max[0], nz_max[1], alpha=0.2)
        # plt.plot(flux_total)
        # plt.plot(flux_total * flux_correction, 'g')
        # ax2 = plt.gca().twinx()
        # ax2.plot(coverage_total)
        #
        # plt.show()
        pack = flux_total_c, colids, nz_max_slice, nz_some_slice
        # pack = flux_total_c
        totals.append(pack)

    return totals[0]


def compute_centroid(rssdata, fiberconf, c1, c2, point, logger=None):
    """Compute centroid near coordinates given by 'point'"""

    logger.debug("LCB configuration is %s", fiberconf.conf_id)

    fibers = fiberconf.conected_fibers(valid_only=True)
    grid_coords = []
    for fiber in fibers:
        grid_coords.append((fiber.x, fiber.y))
    # setup kdtree for searching
    kdtree = KDTree(grid_coords)

    # Other posibility is
    # query using radius instead
    # radius = 1.2
    # kdtree.query_ball_point(points, k=7, r=radius)

    npoints = 19
    # 1 + 6  for first ring
    # 1 + 6  + 12  for second ring
    # 1 + 6  + 12  + 18 for third ring
    points = [point]
    dis_p, idx_p = kdtree.query(points, k=npoints)

    logger.info('Using %d nearest fibers', npoints)
    for diss, idxs, point in zip(dis_p, idx_p, points):
        # For each point
        logger.info('For point %s', point)
        colids = []
        coords = []
        for dis, idx in zip(diss, idxs):
            fiber = fibers[idx]
            colids.append(fiber.fibid - 1)
            coords.append((fiber.x, fiber.y))

        coords = numpy.asarray(coords)
        flux_per_cell = rssdata[colids, c1:c2].mean(axis=1)
        flux_per_cell_total = flux_per_cell.sum()
        flux_per_cell_norm = flux_per_cell / flux_per_cell_total
        # centroid
        scf = coords.T * flux_per_cell_norm
        centroid = scf.sum(axis=1)
        logger.info('centroid: %s', centroid)
        # central coords
        c_coords = coords - centroid
        scf0 = scf - centroid[:, numpy.newaxis] * flux_per_cell_norm
        mc2 = numpy.dot(scf0, c_coords)
        logger.info('2nd order moments, x2=%f, y2=%f, xy=%f', mc2[0,0], mc2[1,1], mc2[0,1])
        return centroid


def compute_dar(img, datamodel, logger=None, debug_plot=False):
    """Compute Diferencial Atmospheric Refraction"""

    fiberconf = datamodel.get_fiberconf(img)
    wlcalib = astropy.wcs.WCS(img[0].header)

    rssdata = img[0].data
    cut1 = 500
    cut2 = 3500
    colids = []
    x = []
    y = []
    for fiber in fiberconf.fibers.values():
        colids.append(fiber.fibid - 1)
        x.append(fiber.x)
        y.append(fiber.y)

    cols = []
    xdar = []
    ydar = []
    delt = 50

    point = [2.0, 2.0]
    # Start in center of range
    ccenter = (cut2 + cut1) // 2
    # Left
    for c in range(ccenter, cut1, -delt):
        c1 = c - delt // 2
        c2 = c + delt // 2

        z = rssdata[colids, c1:c2].mean(axis=1)
        centroid = compute_centroid(rssdata, fiberconf, c1, c2, point, logger=logger)
        cols.append(c)
        xdar.append(centroid[0])
        ydar.append(centroid[1])
        point = centroid

    cols.reverse()
    xdar.reverse()
    ydar.reverse()

    point = [2.0, 2.0]
    # Star over
    # Right
    for c in range(ccenter, cut2, delt):
        c1 = c - delt // 2
        c2 = c + delt // 2
        z = rssdata[colids, c1:c2].mean(axis=1)
        centroid = compute_centroid(rssdata, fiberconf, c1, c2, point)
        cols.append(c)
        xdar.append(centroid[0])
        ydar.append(centroid[1])
        point = centroid

    rr = [[col, 0] for col in cols]
    world = wlcalib.wcs_pix2world(rr, 0)

    if debug_plot:
        import matplotlib.pyplot as plt
        import megaradrp.visualization as vis
        import megaradrp.simulation.refraction as r
        from astropy import units as u

        plt.subplots_adjust(hspace=0.5)
        plt.subplot(111)
        ax = plt.gca()
        plt.xlim([-8, 8])
        plt.ylim([-8, 8])
        col = vis.hexplot(ax, x, y, z, cmap=plt.cm.YlOrRd_r)
        plt.title("Fiber map, %s %s" % (c1, c2))
        cb = plt.colorbar(col)
        cb.set_label('counts')
        plt.show()

        zenith_distance = 60 * u.deg
        press = 79993.2 * u.Pa
        rel = 0.013333333
        temp = 11.5 * u.deg_C

        ll = r.differential_p(
                zenith_distance,
                wl=world[:,0] * u.AA,
                wl_reference=4025 * u.AA,
                temperature=temp,
                pressure=press,
                relative_humidity=rel,
        )
        plt.plot(world[:,0], xdar, '*-')
        plt.plot(world[:,0], ydar, '*-')
        plt.plot(world[:, 0], 2.0 * u.arcsec + ll.to(u.arcsec), '-')
        plt.show()

    return world[:, 0], xdar, ydar


def generate_sensitivity(final, spectrum, star_interp, extinc_interp,
                         wl_coverage1, wl_coverage2, sigma=20.0):

        wcsl = astropy.wcs.WCS(final[0].header)

        r1 = numpy.arange(final[0].shape[1])
        r2 = r1 * 0.0
        lm = numpy.array([r1, r2])
        # Values are 0-based
        wavelen_ = wcsl.all_pix2world(lm.T, 0)
        if wcsl.wcs.cunit[0] == u.dimensionless_unscaled:
            # CUNIT is empty, assume Angstroms
            wavelen = wavelen_[:, 0] * u.AA
        else:
            wavelen = wavelen_[:, 0] * wcsl.wcs.cunit[0]

        wavelen_aa = wavelen.to(u.AA)

        airmass = final[0].header['AIRMASS']
        exptime = final[0].header['EXPTIME']

        response_0 = spectrum / exptime
        valid = response_0 > 0
        # In magAB
        # f(Jy) = 3631 * 10^-0.4 mAB

        mag_ref = (star_interp(wavelen_aa) + extinc_interp(wavelen_aa) * airmass) * u.ABmag
        response_1 = mag_ref.to(u.Jy).value

        r0max = response_0.max()
        r1max = response_1.max()
        r0 = response_0 / r0max
        r1 = response_1 / r1max

        pixm1 = wl_coverage1.start
        pixm2 = wl_coverage1.stop
        pixr1 = wl_coverage2.start
        pixr2 = wl_coverage2.stop

        pixlims = {}
        pixlims['PIXLIMR1'] = pixr1 + 1  # Convert to 1-ref
        pixlims['PIXLIMR2'] = pixr2
        pixlims['PIXLIMM1'] = pixm1 + 1  # Convert to 1-ref
        pixlims['PIXLIMM2'] = pixm2

        max_valid = numpy.zeros_like(valid)
        max_valid[wl_coverage1] = True

        partial_valid = numpy.zeros_like(valid)
        partial_valid[wl_coverage2] = True

        valid = numpy.zeros_like(response_0)
        valid[wl_coverage1] = 1

        pixf1, pixf2 = int(math.floor(pixm1 +  2* sigma)), int(math.ceil(pixm2 - 2 * sigma))

        pixlims['PIXLIMF1'] = pixf1 + 1
        pixlims['PIXLIMF2'] = pixf2

        flux_valid = numpy.zeros_like(valid, dtype='bool')
        flux_valid[pixf1:pixf2] = True

        r0_ens = gaussian_filter(r0, sigma=sigma)

        ratio2 = r0_ens / r1
        s_response = ratio2 * (r0max / r1max)

        # FIXME: add history
        sens = fits.PrimaryHDU(s_response, header=final[0].header)
        # delete second axis keywords
        for key in ['CRPIX2', 'CRVAL2', 'CDELT2', 'CTYPE2']:
            if key in sens.header:
                del sens.header[key]

        sens.header['uuid'] = str(uuid.uuid1())
        sens.header['tunit'] = ('Jy', "Final units")

        update_flux_limits(sens.header, pixlims, wcs=wcsl, ref=0)

        return sens


def subtract_sky(img, datamodel, ignored_sky_bundles=None, logger=None):
    # Sky subtraction

    if logger is None:
        logger = logging.getLogger(__name__)

    logger.info('obtain fiber information')
    sky_img = copy_img(img)
    final_img = copy_img(img)
    fiberconf = datamodel.get_fiberconf(sky_img)
    # Sky fibers
    skyfibs = fiberconf.sky_fibers(valid_only=True,
                                   ignored_bundles=ignored_sky_bundles)
    logger.debug('sky fibers are: %s', skyfibs)
    # Create empty sky_data
    target_data = img[0].data

    target_map = img['WLMAP'].data
    sky_data = numpy.zeros_like(img[0].data)
    sky_map = numpy.zeros_like(img['WLMAP'].data)
    sky_img[0].data = sky_data

    for fibid in skyfibs:
        rowid = fibid - 1
        sky_data[rowid] = target_data[rowid]
        sky_map[rowid] = target_map[rowid]
    # Sum
    coldata = sky_data.sum(axis=0)
    colsum = sky_map.sum(axis=0)

    # Divide only where map is > 0
    mask = colsum > 0
    avg_sky = numpy.zeros_like(coldata)
    avg_sky[mask] = coldata[mask] / colsum[mask]

    # This should be done only on valid fibers
    # The information of which fiber is valid
    # is in the tracemap, not in the header
    for fibid in fiberconf.valid_fibers():
        rowid = fibid - 1
        final_img[0].data[rowid, mask] = img[0].data[rowid, mask] - avg_sky[mask]

    return final_img, img, sky_img



#
# Copyright 2011-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""The MEGARA Data Reduction Pipeline."""

import logging


__version__ = '0.9.3'


# Top level NullHandler
logging.getLogger("megaradrp").addHandler(logging.NullHandler())



#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

""" Twilight fiber flat Calibration Recipes for Megara"""

from __future__ import division, print_function


import numpy
from astropy.io import fits

from numina.core import Result, Parameter
from numina.core.requirements import ObservationResultRequirement
from numina.util.flow import SerialFlow
from numina.exceptions import ValidationError

import megaradrp.requirements as reqs
from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.types import MasterTwilightFlat
from megaradrp.types import ProcessedRSS, ProcessedFrame
# Flat 2D
from megaradrp.processing.combine import basic_processing_with_combination
from numina.array import combine
# Create RSS
from megaradrp.processing.aperture import ApertureExtractor
from megaradrp.processing.wavecalibration import WavelengthCalibrator
from megaradrp.processing.fiberflat import Splitter, FlipLR, FiberFlatCorrector

import numina.core.recipeinout as recipeio
from numina.core.metarecipes import generate_docs


def pixel_2d_check(value):
    """Check this is a valid 2D pixel list"""
    if len(value) != 2:
        raise ValidationError("must have 2 elements")

    return value


def pixel_2d_check_or_none(value):
    """Check this is a valid 2D pixel list or None"""
    if value is None:
        return value
    return pixel_2d_check(value)


@generate_docs
class RecipeInput(recipeio.RecipeInput):
    obresult = ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_bpm = reqs.MasterBPMRequirement()
    master_slitflat = reqs.MasterSlitFlatRequirement()
    master_traces = reqs.MasterAperturesRequirement()
    extraction_offset = Parameter([0.0], 'Offset traces for extraction', accept_scalar=True)
    normalize_region = Parameter([1900, 2100], 'Region used to normalize the flat-field',
                                 validator=pixel_2d_check)
    continuum_region = Parameter([1900, 1900], 'Subtract this region before normalize the flat-field',
                                 validator=pixel_2d_check_or_none)
    master_wlcalib = reqs.WavelengthCalibrationRequirement()
    master_fiberflat = reqs.MasterFiberFlatRequirement()


@generate_docs
class RecipeResult(recipeio.RecipeResult):
    reduced_image = Result(ProcessedFrame)
    reduced_rss = Result(ProcessedRSS)
    master_twilightflat = Result(MasterTwilightFlat)


@recipeio.define_input(RecipeInput)
@recipeio.define_result(RecipeResult)
class TwilightFiberFlatRecipe(MegaraBaseRecipe):
    """Process TWILIGHT_FLAT images and create MASTER_TWILIGHT_FLAT product.

    This recipe process a set of continuum flat images obtained in
    **Twilight Fiber Flat** mode and returns the master twilight flat product
    The recipe also returns the result of processing the input images up to
    slitflat correction. and the result RSS of the processing
    up to wavelength calibration.

    See Also
    --------
    megaradrp.types.MasterTwilightFlat: description of MasterTwilightFlat product

    Notes
    -----
    Images provided in `obresult` are trimmed and corrected from overscan,
    bad pixel mask (if `master_bpm` is not None), bias and dark current
    (if `master_dark` is not None) and corrected from pixel-to-pixel flat
    if `master_slitflat` is not None.
    Images thus corrected are the stacked using the median.

    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces` and resampled according to the wavelength calibration in
    `master_wlcalib`. Then is divided by the `master_fiberflat`.
    The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    To normalize the `master_twilight_flat`, each fiber is divided by the average
    of the column range given in `normalize_region`. This RSS
    image is returned in the field `master_twilightflat` of the recipe result.

    """

    def process_flat2d(self, rinput):
        flow = self.init_filters(rinput, rinput.obresult.configuration)
        final_image = basic_processing_with_combination(
            rinput, flow, method=self.combine_median_scaled
        )
        hdr = final_image[0].header
        self.set_base_headers(hdr)
        return final_image

    def run_reduction_1d(self, img, tracemap, wlcalib, fiberflat, offset=None):
        # 1D, extraction, Wl calibration, Flat fielding
        correctors = []
        correctors.append(ApertureExtractor(tracemap, self.datamodel, offset=offset))
        correctors.append(FlipLR())
        correctors.append(WavelengthCalibrator(wlcalib, self.datamodel))
        correctors.append(FiberFlatCorrector(fiberflat.open(), self.datamodel))

        flow_1d = SerialFlow(correctors)

        reduced_rss =  flow_1d(img)
        return reduced_rss

    def run(self, rinput):
        # Basic processing

        self.logger.info('twilight fiber flat reduction started')

        img = self.process_flat2d(rinput)
        # Copy image
        reduced_image = fits.HDUList([hdu.copy() for hdu in img])
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        reduced_rss = self.run_reduction_1d(img,
                                            rinput.master_traces,
                                            rinput.master_wlcalib,
                                            rinput.master_fiberflat,
                                            offset=rinput.extraction_offset
                                            )

        self.save_intermediate_img(reduced_rss, 'reduced_rss.fits')
        # Measure values in final
        rss_wl_data = reduced_rss[0].data
        mask = self.good_ids_mask(rinput.master_wlcalib)

        start, end = rinput.normalize_region
        self.logger.info('doing mean between columns %d-%d', start, end)
        colapse = rss_wl_data[:, start:end].mean(axis=1)
        if rinput.continuum_region is not None:
            start_c, end_c = rinput.continuum_region
            if end_c > start_c:
                self.logger.info('subtract mean of columns %d-%d', start_c, end_c)
                colapse_c = rss_wl_data[:, start_c:end_c].mean(axis=1)
                colapse -= colapse_c

        # Normalize the colapsed array
        colapse_good = colapse[mask]
        colapse_norm = colapse / colapse_good.mean()
        normalized = numpy.tile(colapse_norm[:, numpy.newaxis], rss_wl_data.shape[1])

        master_t = fits.HDUList([hdu.copy() for hdu in reduced_rss])
        master_t[0].data = normalized
        self.set_base_headers(master_t[0].header)

        self.logger.info('twilight fiber flat reduction ended')
        result = self.create_result(reduced_image=reduced_image,
                                    reduced_rss=reduced_rss,
                                    master_twilightflat=master_t
                                    )

        return result

    def good_ids_mask(self, calibration):

        # Bad fibers, join:
        bad_fibers = calibration.missing_fibers
        bad_fibers.extend(calibration.error_fitting)

        bad_idxs = [fibid - 1 for fibid in bad_fibers]

        good_idxs_mask = numpy.ones((calibration.total_fibers,), dtype='bool')
        good_idxs_mask[bad_idxs] = False
        return good_idxs_mask

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(TwilightFiberFlatRecipe, self).set_base_headers(hdr)
        hdr['NUMTYPE'] = ('MasterTwilightFlat', 'Product type')
        return hdr

    def combine_median_scaled(self, arrays, masks=None, dtype=None, out=None,
                                    zeros=None, scales=None,
                                    weights=None):

        median_vals = numpy.array([numpy.median(arr) for arr in arrays])
        self.logger.info("median values are %s", median_vals)
        # normalize by max value
        median_max = numpy.max(median_vals)
        scales = median_max / median_vals
        self.logger.info("scale values are %s", scales)
        return combine.median(arrays, scales=scales, dtype=dtype)



#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import math

from astropy import units as u

from .efficiency import Efficiency


class LightFiber(object):
    def __init__(self, name, fibid, transmission=None, inactive=False):
        self.name = name
        self.fibid = fibid
        # Geometry of the fibers
        self.size = 0.31 * u.arcsec
        self.area = math.sqrt(3) * self.size ** 2 / 2.0
        self.fwhm = 3.6
        self.sigma = self.fwhm / 2.3548
        self.inactive = inactive

        if transmission is None:
            self._transmission = Efficiency()
        else:
            self._transmission = transmission

    def transmission(self, wl):
        return self._transmission.response(wl)

    def config_info(self):
        return {'name': self.name,
                'fibid': self.fibid,
                'inactive': self.inactive
                }


class FiberSet(object):
    def __init__(self, name, size, fwhm):
        self.fibers = {}
        self.bundles = {}
        self.name = name

        # Geometry of the fibers
        self.size = size
        self.area = math.sqrt(3) * self.size ** 2 / 2.0
        self.fwhm = fwhm
        self.sigma = self.fwhm / 2.3548

    @property
    def nfibers(self):
        return len(self.fibers)

    def transmission(self, wlin):
        # Loop over fibers
        import numpy

        result = numpy.zeros((self.nfibers, wlin.shape[0]))
        for lf in self.fibers.values():
            result[lf.fibid - 1] = lf.transmission(wlin)
        return result

    def config_info(self):
        return {'name': self.name,
                'nfibers': self.nfibers,
                'fwhm': self.fwhm,
                'sigma': self.sigma,
                }


#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import math

from numina.instrument.hwdevice import HWDevice
from numina.instrument.simulation.efficiency import Efficiency

from megaradrp.instrument.components.instrument import FocusActuator


class M2FocusActuator(FocusActuator):
    def __init__(self, name, parent=None):
        super(M2FocusActuator, self).__init__(name, parent)

        # Focus
        self.internal_focus_factor = 1.0
        self._ref_focus = 1120
        self._internal_focus = self._ref_focus

    def set_focus(self, x):
        """Arbitrary parametrization of the focus"""
        if x < -4000 or x > 4000:
            raise ValueError('telescope focus out of limits')

        self.internal_focus_factor = 1+ 1.9 * (math.cosh((x - self._ref_focus) / 3000.0) - 1)
        self._internal_focus = x


class Telescope(HWDevice):

    def __init__(self, name, diameter, transmission=None):
        super(Telescope, self).__init__(name)

        self.diameter = diameter
        self.area = 0.25 * math.pi * diameter**2
        self.inc = None

        if transmission is None:
            self._transmission = Efficiency()
        else:
            self._transmission = transmission

        self.focus_actuator = M2FocusActuator('Focus', self)

    def transmission(self, wl):
        return self._transmission.response(wl)

    def connect(self, atmmodel):
        self.inc = atmmodel

    def set_focus(self, x):
        """Arbitrary parametrization of the focus"""

        self.focus_actuator.set_focus(x)


#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Fiber flat calibration Recipe for Megara"""

from __future__ import division, print_function

import numpy
from astropy.io import fits
import matplotlib.pyplot as plt
from numina.core import Result, Parameter
import numina.exceptions
from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.types import MasterFiberFlat
import megaradrp.requirements as reqs
from megaradrp.types import ProcessedRSS, ProcessedFrame

# Flat 2D
from megaradrp.processing.combine import basic_processing_with_combination
from numina.array import combine
# Create RSS
from megaradrp.processing.aperture import ApertureExtractor
from megaradrp.processing.wavecalibration import WavelengthCalibrator
from megaradrp.processing.fiberflat import Splitter, FlipLR


def _smoothing_window_check(val):
    """Check 'smoothing_window' """
    #raise ValueError("something")
    if val < 5:
        raise numina.exceptions.ValidationError("must be >= 5")
    if val % 2 == 0:
        raise numina.exceptions.ValidationError("must be odd")
    return val


class FiberFlatRecipe(MegaraBaseRecipe):
    """Process FIBER_FLAT images and create MASTER_FIBER_FLAT product.

    This recipe process a set of continuum flat images obtained in
    **Fiber Flat** mode and returns the master fiber flat product
    The recipe also returns the result of processing the input images up to
    slitflat correction. and the result RSS of the processing
    up to wavelength calibration.

    See Also
    --------
    megaradrp.products.MasterFiberFlat: description of MasterFiberFlat product
    megaradrp.processing.aperture: aperture extraction
    megaradrp.processing.wavecalibration: resampling for wavelength calibration

    Notes
    -----
    Images provided in `obresult` are trimmed and corrected from overscan,
    bad pixel mask (if `master_bpm` is not None), bias and dark current
    (if `master_dark` is not None) and corrected from pixel-to-pixel flat
    if `master_slitflat` is not None.
    Images thus corrected are the stacked using the median.

    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces` and resampled accoding to the wavelength calibration in
    `master_wlcalib`. The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    To normalize the `master_fiberflat`, each fiber is divided by a smoothed
    version (using a Savitzky-Golay filter) of the average of the valid fibers.
    Finally, all the pixels with information are fiiled with ones. This RSS
    image is returned in the field `master_fiberflat` of the recipe result.

    """

    # Requirements
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_bpm = reqs.MasterBPMRequirement()
    master_slitflat = reqs.MasterSlitFlatRequirement()
    master_traces = reqs.MasterAperturesRequirement()
    smoothing_window = Parameter(31, 'Window for smoothing (must be odd)',
                                 validator=_smoothing_window_check
                                 )
    extraction_offset = Parameter([0.0], 'Offset traces for extraction', accept_scalar=True)
    master_wlcalib = reqs.WavelengthCalibrationRequirement()

    # Results
    reduced_image = Result(ProcessedFrame)
    reduced_rss = Result(ProcessedRSS)
    master_fiberflat = Result(MasterFiberFlat)

    def process_flat2d(self, rinput):
        flow = self.init_filters(rinput, rinput.obresult.configuration)
        final_image = basic_processing_with_combination(rinput, flow, method=combine.median)
        hdr = final_image[0].header
        self.set_base_headers(hdr)
        return final_image

    def obtain_fiber_flat(self, rss_wl, wlcalib, col1=1900, col2=2100, window=31, degree=3):
        from scipy.signal import savgol_filter
        from scipy.interpolate import UnivariateSpline

        # Bad fibers, join:
        bad_fibers = wlcalib.missing_fibers
        bad_fibers.extend(wlcalib.error_fitting)
        # print(bad_fibers)
        bad_idxs = [fibid - 1 for fibid in bad_fibers]
        # print(bad_idxs)

        good_idxs_mask = numpy.ones((wlcalib.total_fibers,), dtype='bool')
        good_idxs_mask[bad_idxs] = False

        # Collapse all fiber spectrum
        xcol = slice(col1, col2)

        data0 = rss_wl[0].data

        col_mean = data0[:, xcol].mean(axis=1)
        # Filter positive values and valid fibers
        col_mean_pos = (col_mean > 0)
        valid_mask = col_mean_pos & good_idxs_mask

        col_good_mean = col_mean[valid_mask]

        data_good = data0[valid_mask] / col_good_mean[:, numpy.newaxis]
        data_good[numpy.isnan(data_good)] = 0.0

        # Crappy way
        # This extension was created by WLcalibrator
        wlmap = rss_wl['WLMAP'].data
        mm = numpy.sum(wlmap, axis=0)
        # skip 0 in divisions
        mask_noinfo = mm < 1
        mm[mask_noinfo] = 1
        # Filter collapse to smooth it
        collapse = numpy.sum(data_good, axis=0) / mm
        # Smooting works bad very near the border (overshooting)
        collapse_smooth = savgol_filter(collapse, window, degree)
        collapse_smooth[mask_noinfo] = 1.0

        xx = numpy.arange(collapse.shape[0])
        xx_v = xx[~mask_noinfo]
        collapse_v = collapse[~mask_noinfo]

        # Spline degree (3 or 5)
        degree_s = 5
        interpol_univ = UnivariateSpline(xx_v, collapse_v, k=degree_s)
        collapse_smooth_s = collapse_smooth.copy()
        collapse_smooth_s[~mask_noinfo] = interpol_univ(xx_v)
        collapse_smooth_s[mask_noinfo] = 1.0

        if self.intermediate_results:
            numpy.savetxt('collapse.txt', collapse)
            numpy.savetxt('mask_noinfo.txt', mask_noinfo)
            fig, ax = plt.subplots()
            ax.plot(xx, collapse, '.', label='collapsed')
            ax.plot(xx, collapse_smooth, '-', label='savgol{}'.format(degree))
            ax.plot(xx, collapse_smooth_s, '--', label='spline{}'.format(degree_s))
            ax.legend()
            plt.savefig('collapsed_smooth.png')
            plt.close()

        # Divide each fiber in rss_wl by spectrum
        gmean = col_good_mean.mean()
        data1 = rss_wl[0].data / collapse_smooth_s
        data1 /= gmean
        # Fill values with ones to avoid NaNs
        data2 = numpy.where(wlmap > 0, data1, 1.0)

        self.logger.warning("Copy all extensions for the moment")
        rss_wl2 = fits.HDUList([hdu.copy() for hdu in rss_wl])
        rss_wl2[0].data = data2
        return rss_wl2

    def run(self, rinput):
        """Execute the recipe.

        Parameters
        ----------
        rinput : RecipeInput

        Returns
        -------
        RecipeResult

        """

        img = self.process_flat2d(rinput)
        self.save_intermediate_img(img, 'reduced_image.fits')
        splitter1 = Splitter()
        calibrator_aper = ApertureExtractor(
            rinput.master_traces,
            self.datamodel,
            offset=rinput.extraction_offset
        )
        splitter2 = Splitter()
        calibrator_wl = WavelengthCalibrator(rinput.master_wlcalib, self.datamodel)
        flipcor = FlipLR()

        img = splitter1(img)
        flat2d = splitter1.out # Copy before extraction
        img = calibrator_aper(img)
        img = splitter2(img)
        rss_base = splitter2.out # Copy before el calibration
        self.logger.debug('Flip RSS left-rigtht, before WL calibration')
        img = flipcor.run(img)
        # Calibrate in WL
        rss_wl = calibrator_wl(img)
        self.save_intermediate_img(rss_wl, 'reduced_rss.fits')

        # Obtain flat field
        self.logger.info('Normalize flat field')
        rss_wl2 = self.obtain_fiber_flat(rss_wl, rinput.master_wlcalib, window=rinput.smoothing_window)
        rss_wl2[0].header = self.set_base_headers(rss_wl2[0].header)
        result = self.create_result(
            master_fiberflat=rss_wl2,
            reduced_image=flat2d,
            reduced_rss=rss_base
        )
        return result

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(FiberFlatRecipe, self).set_base_headers(hdr)
        hdr['NUMTYPE'] = ('MasterFiberFlat', 'Product type')
        return hdr


#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

""" Trace model recipe for Megara"""

from __future__ import division, print_function

import math
import bisect
import multiprocessing as mp

import numpy as np
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
from astropy.modeling import fitting
from astropy.modeling.functional_models import Const1D
from numina.core import Result, Requirement, Parameter
from numina.array import combine
from numina.modeling.gaussbox import GaussBox, gauss_box_model
from numina.frame.utils import copy_img

from megaradrp.products.modelmap import ModelMap
from megaradrp.products.modelmap import GeometricModel
from megaradrp.processing.aperture import ApertureExtractor
from megaradrp.types import ProcessedImage, ProcessedRSS
from megaradrp.processing.combine import basic_processing_with_combination
from megaradrp.core.recipe import MegaraBaseRecipe
import megaradrp.requirements as reqs


class ModelMapRecipe(MegaraBaseRecipe):
    """Provides fiber profile information from continuum flat images.

    This recipe process a set of continuum flat images obtained in
    *Trace Map* mode and returns the fiber profile information required
    to perform advanced fiber extraction in other recipes. The recipe also
    returns the result of processing the input images upto dark correction.

    See Also
    --------
    megaradrp.products.modelmap.ModelMap: description of ModelMap product
    megaradrp.recipes.calibration.tracemap.TraceMapRecipe: description of TraceMap recipe
    megaradrp.instrument.configs: instrument configuration

    Notes
    -----
    Images provided in `obresult` are trimmed and corrected from overscan,
    bad pixel mask (if `master_bpm` is not None), bias and dark current
    (if `master_dark` is not None).
    Images thus corrected are the stacked using the median.

    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result and will be used for
    fiting the profiles of the fibers.

    The approximate central position of the fibers is obtained from
    `master_traces`. Then, for each 100 columns of the reduced image, a vertical
    cut in the image is fitted to a sum of fiber profiles, being the profile
    a gaussian convolved with a square.

    The fits are made in parallel, being the number of processes controlled
    by the parameter `processes`, with the default value of 0 meaning to use
    the number of cores minus 2 if the number of cores is greater or equal to 4,
    one process otherwise.

    After the columns are fitted, the profiles (central position and sigma)
    are interpolated to all columns using splines. The coefficientes of the
    splines are stored in the final `master_traces` object.

    The recipe returns also the RSS obtained by applying advanced extraction
    to `reduced_image`. As an intermediate result, the recipe proceduces DS9
    reg files with the position of the center of the profiles, that can be
    used with raw and reduced images.

    """

    # Requirements
    master_bpm = reqs.MasterBPMRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_slitflat = reqs.MasterSlitFlatRequirement()
    # FIXME: this is not really necessary, it can be computed
    # from the data
    master_traces = reqs.MasterTraceMapRequirement()
    processes = Parameter(0, 'Number of processes used for fitting')
    debug_plot = Parameter(0, 'Save intermediate tracing plots')
    # Results
    reduced_image = Result(ProcessedImage)
    reduced_rss = Result(ProcessedRSS)
    master_model = Result(ModelMap)

    def run(self, rinput):

        self.logger.info('starting model map recipe')

        obresult = rinput.obresult
        obresult_meta = obresult.metadata_with(self.datamodel)
        if rinput.processes == 0:
            have = mp.cpu_count()
            if have >= 4:
                processes = mp.cpu_count() - 2
            else:
                processes = 1
        else:
            processes = rinput.processes
        self.logger.debug('using %d processes', processes)

        self.logger.info('start basic reduction')
        flow1 = self.init_filters(rinput, rinput.obresult.configuration)
        reduced = basic_processing_with_combination(rinput, flow1, method=combine.median)
        self.set_base_headers(reduced[0].header)
        self.logger.info('end basic reduction')

        self.save_intermediate_img(reduced, 'reduced_image.fits')

        self.logger.debug('create result model')

        # insconf = obresult.configuration

        # boxes = insconf.get_property('pseudoslit.boxes')
        # values = insconf.get_property('pseudoslit.boxes_positions')
        # cstart0 = values['ref_column']
        # box_borders0 = values['positions']

        # box_borders, cstart = self.refine_boxes_from_image(reduced, box_borders0, cstart0)

        # self.logger.debug("original boxes: %s", box_borders0)
        # self.logger.debug("refined boxes: %s", box_borders)

        model_map = ModelMap(instrument=obresult.instrument)

        self.logger.debug('update metadata in model')
        model_map.update_metadata(self)
        fiberconf = self.datamodel.get_fiberconf(reduced)
        model_map.total_fibers = fiberconf.nfibers
        model_map.missing_fibers = rinput.master_traces.missing_fibers
        model_map.tags = self.extract_tags_from_ref(reduced, model_map.tag_names(), base=obresult.labels)
        # model_map.boxes_positions = box_borders
        # model_map.ref_column = cstart
        model_map.update_metadata(self)
        model_map.update_metadata_origin(obresult_meta)
        # Temperature in Celsius with 2 decimals
        model_map.tags['temp'] = round(obresult_meta['info'][0]['temp'] - 273.15, 2)

        self.logger.info('perform model fitting')

        tracemap = rinput.master_traces
        data = reduced[0].data
        sigma = 1.53
        cols = range(100, 4100, 100)
        # cols = range(100, 200, 100)
        valid = [f.fibid for f in tracemap.contents if f.valid]
        ncol = tracemap.total_fibers
        nrow = data.shape[0]
        nfit = data.shape[1]
        results_get = fit_model(data, tracemap, valid, nrow, ncol, sigma, cols, processes)

        self.logger.info('perform model fitting end')

        self.logger.info('interpolate parameters')

        array_mean = np.zeros((ncol, nfit))
        array_std = np.zeros((ncol, nfit))

        xcol = np.arange(nfit)

        mean_splines = {}
        std_splines = {}

        for fibid in valid:
            g_amp = []
            g_std = []
            g_mean = []
            g_col = []

            dolog = (fibid % 100 == 0)

            if dolog:
                self.logger.debug('compute fibid %d', fibid)

            for calc_col, vals in results_get:
                param = vals[fibid]
                g_col.append(calc_col)
                g_std.append(param['stddev'])
                g_mean.append(param['mean'])
                g_amp.append(param['amplitude'])

            interpol_std = UnivariateSpline(g_col, g_std, k=5)
            interpol_mean = UnivariateSpline(g_col, g_mean, k=3)

            if self.intermediate_results:
                if dolog:
                    self.logger.debug('saving plots')
                plt.title('std fib{:03d}'.format(fibid))
                plt.plot(g_col, g_std, 'b*')
                plt.plot(g_col, interpol_std(g_col), 'r')
                plt.savefig('fib_{:03d}_std.png'.format(fibid))
                plt.close()
                plt.title('mean fin{:03d}'.format(fibid))
                plt.plot(g_col, g_mean, 'b*')
                plt.plot(g_col, interpol_mean(g_col), 'r')
                plt.savefig('fib_{:03d}_mean.png'.format(fibid))
                plt.close()
                if dolog:
                    self.logger.debug('saving plots end')

            mean_splines[fibid] = interpol_mean
            std_splines[fibid] = interpol_std

            row = fibid - 1
            array_mean[row] = interpol_mean(xcol)
            array_std[row] = interpol_std(xcol)

            params = {'stddev': interpol_std, 'mean': interpol_mean}
            model = {'model_name': 'gaussbox', 'params': params}
            # if invalid. missing, model = {}
            m = GeometricModel(
                fibid,
                boxid=1, # FIXME: not counting this
                start=1,
                stop=nfit,
                model=model
            )

            model_map.contents.append(m)

        self.logger.info('interpolate parameters end')

        # perform extraction with our own calibration
        self.logger.info('perform extraction with computed calibration')
        calibrator_aper = ApertureExtractor(model_map, self.datamodel)
        reduced_copy = copy_img(reduced)
        reduced_rss = calibrator_aper(reduced_copy)

        if self.intermediate_results:
            with open('ds9.reg', 'w') as ds9reg:
                model_map.to_ds9_reg(ds9reg, rawimage=False,
                                     numpix=100, fibid_at=2048)

            with open('ds9_raw.reg', 'w') as ds9reg:
                model_map.to_ds9_reg(ds9reg, rawimage=True,
                                     numpix=100, fibid_at=2048)

        self.logger.info('ending model map recipe')
        result = self.create_result(reduced_image=reduced,
                                    reduced_rss=reduced_rss,
                                    master_model=model_map)

        return result


def initial_base(boxd1d, ecenters, npix=5):
    nfib = len(ecenters)

    offset = npix // 2
    yfit = np.empty((npix, nfib))
    xfit = np.arange(npix) - offset

    for i in range(npix):
        yfit[i, :] = boxd1d[ecenters + (i - offset)]

    coeff = np.polyfit(xfit, np.log(yfit), deg=2)
    c, b, a = coeff

    sig2 = -1 / (2 * c)
    mu = b / sig2
    ampl = np.exp(a - mu ** 2)

    return ampl, mu, sig2


def calc1d_N(boxd1d, valid, centers1d, sigma, lateral=0, reject=3, nloop=1,
             fixed_centers=False, init_simple=False):

    ecenters = np.ceil(centers1d - 0.5).astype('int')
    nfib = len(centers1d)

    # FIXME, hardcoded
    xl = np.arange(4112.0)

    if init_simple:
        ampl, mu, sig2 = initial_base(boxd1d, ecenters, npix=5)
        sig_calc = np.sqrt(sig2)
    else:
        sig_calc = sigma * np.ones((nfib,))
        ampl = [boxd1d[ecenters[i]] / 0.25 for i in range(nfib)]

    init = {}
    for i in range(nfib):
        fibid = valid[i]
        init[fibid] = {}
        init[fibid]['mean'] = centers1d[i]
        init[fibid]['stddev'] = sig_calc[i]
        init[fibid]['amplitude'] = ampl[i]

    # plt.plot(sig_calc / sigma)
    # plt.show()

    # total fits is 2 * lateral + 1
    total = reject + lateral
    yl = boxd1d

    for il in range(nloop):

        # print 'loop', il, datetime.datetime.now()

        # permutation of the valid fibers
        values = np.random.permutation(valid)

        for val in values:
            #

            m1 = max(0, int(math.ceil(init[val]['mean'] - 0.5)) - 6 * total)
            m2 = int(math.ceil(init[val]['mean'] - 0.5)) + 6 * total

            y = yl[m1:m2].copy()
            xt = xl[m1:m2]

            pos = bisect.bisect(valid, val)
            candidates = valid[max(pos - total - 1, 0): pos + total]
            dis_f = [abs(c - val) for c in candidates]
            # for p in range(max(0, val-S), min(nfib, val+S+1)):
            fit_ids = []
            for c, df in zip(candidates, dis_f):
                if df > lateral:
                    # remove contribution
                    y -= gauss_box_model(xt, **init[c])
                else:
                    fit_ids.append(c)

            # num_of_fibers = len(fit_ids)
            # print('lateral', lateral, val, fit_ids)
            # print('num', num_of_fibers)

            # Extract values to create initials
            model = Const1D(amplitude=0.0)
            for fib_id in fit_ids:
                newg = GaussBox(
                    amplitude=init[fib_id]['amplitude'],
                    mean=init[fib_id]['mean'],
                    stddev=init[fib_id]['stddev']
                )
                model = model + newg

            # Set fit conditions
            model.amplitude_0.fixed = True
            for idx, fib_id in enumerate(fit_ids, 1):
                key = "mean_%d" % idx
                m_mean = getattr(model, key)
                if fixed_centers:
                    m_mean.fixed = True
                else:
                    m_mean.min = init[fib_id]['mean'] - 0.5
                    m_mean.max = init[fib_id]['mean'] + 0.5

                key = "stddev_%d" % idx
                m_stddev = getattr(model, key)
                m_stddev.min = init[fib_id]['stddev'] - 0.5
                m_stddev.max = init[fib_id]['stddev'] + 0.5

                key = "hpix_%d" % idx
                m_hpix = getattr(model, key)
                m_hpix.fixed = True

            fitter = fitting.LevMarLSQFitter()
            model_fitted = fitter(model, xt, y)

            # Extract values
            # This extracts only val
            # val_idx = fit_ids.index(val) + 1
            # na = getattr(model_fitted, 'amplitude_%d' % val_idx).value
            # nm = getattr(model_fitted, 'mean_%d' % val_idx).value
            # ns = getattr(model_fitted, 'stddev_%d' % val_idx).value

            # init[val]['amplitude'] = na
            # init[val]['mean'] = nm
            # init[val]['stddev'] = ns

            for val_idx, fibid in enumerate(fit_ids, 1):
                # This will overwrite fits with fits performed later
                # if fibid == val:
                na = getattr(model_fitted, 'amplitude_%d' % val_idx).value
                nm = getattr(model_fitted, 'mean_%d' % val_idx).value
                ns = getattr(model_fitted, 'stddev_%d' % val_idx).value

                init[fibid]['amplitude'] = na
                init[fibid]['mean'] = nm
                init[fibid]['stddev'] = ns
                # break

    return init


def calc2_base(column, centers, valid, sigma, nloop=10, do_plot=False):
    scale = column.max()
    column_norm = column / scale
    final = calc1d_N(column_norm, valid, centers, sigma, lateral=2, nloop=nloop)

    for idx, params in final.items():
        final[idx]['amplitude'] *= scale

    if do_plot:
        pass # calc2_plot(final, centers, valid, sigma)

    return final


def calc_para_2(data, calc_col, tracemap, valid, nrow, ncol, sigma,
                nloop=10, average=0, calc_init=True):
    if average > 0:
        yl = data[:, calc_col - average:calc_col - average + 1].mean(axis=1)
    else:
        yl = data[:, calc_col]

    centers = np.array([f.polynomial(calc_col) for f in tracemap.contents if f.valid])
    print ('calc_col is', calc_col)

    final = calc2_base(yl, centers, valid, sigma, nloop=nloop, do_plot=False)

    return calc_col, final


def fit_model(data, tracemap, valid, nrow, ncol, sigma, cols, processes=20):

    pool = mp.Pool(processes)

    results = [pool.apply_async(
        calc_para_2,
        args=(data, col, tracemap, valid, nrow, ncol, sigma),
        kwds={'nloop': 3, 'average': 2}
    ) for col in cols]

    results_get = [p.get() for p in results]
    return results_get



#
# Copyright 2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Custom tables for MEGARA"""


from sqlalchemy import Integer, String, DateTime, Float
from sqlalchemy import CHAR
from sqlalchemy import Table, Column, ForeignKey
from sqlalchemy.orm import relationship, synonym


from numinadb.base import Base


class MegaraFrame(Base):
    __tablename__ = 'megara_frames'
    id = Column(Integer, primary_key=True)
    uuid = Column(CHAR(32), nullable=True)
    name = Column(String(100), nullable=False, unique=True)
    # name = Column(String(100), unique=True, nullable=False)
    ob_id = Column(String,  ForeignKey("obs.id"), nullable=False)
    object = Column(String)
    start_time = Column(DateTime)
    exposure_time = Column(Float)
    insmode = Column(String)
    vph = Column(String)
    completion_time = Column(DateTime)







#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import print_function, division

import logging
import datetime

import numpy as np
import numpy
from scipy.interpolate import LSQUnivariateSpline
from astropy.io import fits


from numina.processing import Corrector


_logger = logging.getLogger(__name__)


_direc = ['normal', 'mirror']


def trimOut(img, detconf, direction='normal', out='trimmed.fits'):
    """
    :param data: original data to be trimmed. Can be either a string <path>, an ImageHDU or a numpy array
    :param direction: Can be either 'normal' or 'mirror'
    :param out: When the original data is a path, the returned value is a file too
    :param bins: is the index of the _binning dictionary
    :return:
    """

    if issubclass(str, type(img)):
        with fits.open(img) as hdul:
            hdu = trim_and_o_array(hdul[0].data, detconf, direction=direction)
            fits.writeto(out, hdu, overwrite=True)

    elif isinstance(img, fits.PrimaryHDU):
        finaldata = trim_and_o_array(img.data, detconf, direction=direction)
        img.data = finaldata
        return img

    elif isinstance(img, numpy.ndarray):
        return trim_and_o_array(img, detconf, direction=direction)


def trim_and_o_array(array, detconf, direction='normal'):
    """Trim a MEGARA array with overscan."""

    if direction not in _direc:
        raise ValueError("%s must be either 'normal' or 'mirror'" % direction)

    if direction == 'normal':
        direcfun = lambda x: x
    else:
        direcfun = numpy.fliplr

    trim1 = get_conf_value(detconf, 'trim1')
    trim2 = get_conf_value(detconf, 'trim2')
    bng = get_conf_value(detconf, 'bng')

    if bng not in [[1,1],[1,2],[2,1],[2,2]]:
        raise ValueError("%s must be one if '11', '12', '21, '22'" % bng)

    nr2 = ((trim1[0][1] - trim1[0][0]) + (trim2[0][1]-trim2[0][0])) // bng[0]
    nc2 = (trim1[1][1] - trim1[1][0]) // bng[1]
    nr = (trim1[0][1] - trim1[0][0]) // bng[0]

    trim1[0][0] //= bng[0]
    trim1[0][1] //= bng[0]
    trim2[0][0] //= bng[0]
    trim2[0][1] //= bng[0]

    trim1[1][0] //= bng[1]
    trim1[1][1] //= bng[1]
    trim2[1][0] //= bng[1]
    trim2[1][1] //= bng[1]

    finaldata = numpy.empty((nr2, nc2), dtype='float32')

    finY = trim1[0][1] + trim2[0][0]
    finaldata[:nr, :] = direcfun(array[:trim1[0][1], trim1[1][0]:trim1[1][1]])
    finaldata[nr:, :] = direcfun(array[trim2[0][0]:finY, trim2[1][0]:trim2[1][1]])

    return finaldata


def get_conf_value(confFile, key):
    return confFile[key]


def apextract_weights(data, weights, size=4096):
    """Extract apertures using a map of weights."""

    import multiprocessing as mp

    def decompress(tar_file):
        """
        :param tar_name: <str> name of the tar file
        :return: None
        """

        name = tar_file.fileobj.name.split('.tar')[0]
        aux = tar_file.extractall(name+'/')
        return name


    processes = mp.cpu_count()-2

    path = decompress(weights)

    pool = mp.Pool(processes=processes)
    results = [pool.apply_async(load_files_paralell,
                                args=(ite, path)) for ite in range(size)]
    results = [p.get() for p in results]

    pool2 = mp.Pool(processes=processes)
    extracted_w = [pool2.apply_async(extract_w_paralell,
                                args=(data[:,ite], results[ite])) for ite in range(size)]
    extracted_w = [p.get() for p in extracted_w]

    return numpy.array(extracted_w)


def load_files_paralell(col, path):
    """"
    :param col: <str,int> name of the fits file. It is a counter
    :param path: <str> path where *.npz are
    :return: csr_matrix
    """
    from scipy.sparse import csr_matrix

    filename = '%s/%s.npz' % (path, col)
    loader = numpy.load(filename)
    return csr_matrix(
        (loader['data'], loader['indices'], loader['indptr']),
        shape=loader['shape'])



def extract_w_paralell(img, mlist):
    """
    :param img: <fits>
    :param mlist: <list> one element of the csr_matrix
    :return: <ndarray> result of lsqr
    """

    from scipy.sparse.linalg import lsqr
    x = lsqr(mlist, img)
    return x[0]


class OverscanCorrector(Corrector):
    """A Corrector Node that corrects a MEGARA image from overscan."""

    def __init__(self, detconf, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        trim1 = get_conf_value(detconf, 'trim1')
        trim2 = get_conf_value(detconf, 'trim2')
        bng = get_conf_value(detconf, 'bng')
        overscan1 = get_conf_value(detconf, 'overscan1')
        overscan2 = get_conf_value(detconf, 'overscan2')
        prescan1 = get_conf_value(detconf, 'prescan1')
        prescan2 = get_conf_value(detconf, 'prescan2')
        middle1 = get_conf_value(detconf, 'middle1')
        middle2 = get_conf_value(detconf, 'middle2')

        auxX, auxY, auxZ, auxT = self.data_binning(trim1, bng)
        middleX, middleY, middleZ, middleT = self.data_binning(middle1, bng)
        prescanX, prescanY, prescanZ, prescanT = self.data_binning(prescan1,
                                                                   bng)
        overscanX, overscanY, overscanZ, overscanT = self.data_binning(
            overscan1, bng)
        self.trim1 = (slice(auxX, auxY), slice(auxZ, auxT))
        self.orow1 = (slice(middleX, middleY), slice(middleZ, middleT))
        self.pcol1 = (slice(prescanX, prescanY), slice(prescanZ, prescanT))
        self.ocol1 = (slice(overscanX, overscanY), slice(overscanZ, overscanT))

        auxX, auxY, auxZ, auxT = self.data_binning(trim2, bng)
        middleX, middleY, middleZ, middleT = self.data_binning(middle2, bng)
        prescanX, prescanY, prescanZ, prescanT = self.data_binning(prescan2,
                                                                   bng)
        overscanX, overscanY, overscanZ, overscanT = self.data_binning(
            overscan2, bng)
        self.trim2 = (slice(auxX, auxY), slice(auxZ, auxT))
        self.orow2 = (slice(middleX, middleY), slice(middleZ, middleT))
        self.pcol2 = (slice(prescanX, prescanY), slice(prescanZ, prescanT))
        self.ocol2 = (slice(overscanX, overscanY), slice(overscanZ, overscanT))

        # self.test_image()
        super(OverscanCorrector, self).__init__(datamodel=datamodel,
                                                calibid=calibid,
                                                dtype=dtype)

    def data_binning(self, data, binning):
        '''
         Axis:x --> factorX
         Axis:y --> factorY
        '''
        factorX = 1.0 / binning[1]
        factorY = 1.0 / binning[0]
        x = int(factorY * data[0][0])
        y = int(factorY * data[0][1])
        z = int(factorX * data[1][0])
        t = int(factorX * data[1][1])

        return x, y, z, t

    def test_image(self):
        import astropy.io.fits as fits

        data = np.empty((4212, 4196), dtype='float32')
        data[self.pcol1] += 1
        data[self.orow1] += 10
        data[self.ocol1] += 100
        data[self.trim1] += 1000

        data[self.pcol2] += 5
        data[self.orow2] += 50
        data[self.ocol2] += 500
        data[self.trim2] += 5000

        fits.writeto('eq_estimado.fits', data, overwrite=True)

    def run(self, img):
        imgid = self.get_imgid(img)
        data = img[0].data

        # p1 = data[self.pcol1].mean()
        # _logger.debug('prescan1 is %f', p1)
        # or1 = data[self.orow1].mean()
        # _logger.debug('row overscan1 is %f', or1)
        oc1 = np.median(data[self.ocol1])
        _logger.debug('median col overscan1 is %f', oc1)
        # avg = (p1 + or1 + oc1) / 3.0

        oc2 = np.median(data[self.ocol2])
        _logger.debug('median col overscan2 is %f', oc2)

        _logger.debug('compute spline for overscan1')
        fit1, spl1 = self.eval_spline_amp1(data)
        data[self.trim1] -= fit1[:, numpy.newaxis]

        _logger.debug('compute spline for overscan2')
        fit2, spl2 = self.eval_spline_amp2(data)
        data[self.trim2] -= fit2[:, numpy.newaxis]

        hdr = img['primary'].header
        hdr['NUM-OVPE'] = self.calibid
        hdr['history'] = 'Overscan correction with {}'.format(self.calibid)
        hdr['history'] = 'Overscan correction time {}'.format(datetime.datetime.utcnow().isoformat())
        # hdr['history'] = 'Mean of prescan1 is %f' % p1
        hdr['history'] = 'Median of col overscan1 is {}'.format(oc1)
        hdr['history'] = 'Overscan1 correction is {}'.format('spline3')
        # hdr['history'] = 'prescan2 is %f' %  p2
        hdr['history'] = 'Median of col overscan2 is {}'.format(oc2)
        hdr['history'] = 'Overscan2 correction is {}'.format('spline3')

        for spl, label in zip([spl1, spl2], ['overscan1', 'overscan2']):
            k, c, deg = spl._eval_args
            hdr['history'] = '{} deg {}'.format(label, deg)
            hdr['history'] = '{} knots {}'.format(label, list(k))
            hdr['history'] = '{} coeffs {}'.format(label, list(c))

        return img

    def fit_spline_amp1(self, data):

        region = self.ocol1
        u = numpy.arange(region[0].start, region[0].stop)
        v = data[region].mean(axis=1)
        knots1 = [1200]
        spl1 = LSQUnivariateSpline(u, v, knots1, k=3)
        return spl1

    def eval_spline_amp1(self, data):
        spl1 = self.fit_spline_amp1(data)
        region = self.ocol1
        u = numpy.arange(region[0].start, region[0].stop)
        fit1 = spl1(u)
        return fit1, spl1

    def fit_spline_amp2(self, data):
        region = self.ocol2
        u = numpy.arange(region[0].start, region[0].stop)
        v = data[region].mean(axis=1)
        knots2 = [3100]
        spl2 = LSQUnivariateSpline(u, v, knots2, k=3)
        return spl2

    def eval_spline_amp2(self, data):
        spl2 = self.fit_spline_amp2(data)
        region = self.ocol2
        u = numpy.arange(region[0].start, region[0].stop)
        fit2 = spl2(u)
        return fit2, spl2


class TrimImage(Corrector):
    """A Corrector Node that trims MEGARA images."""

    def __init__(self, detconf, datamodel=None, calibid='calibid-unknown', dtype='float32'):
        self.detconf = detconf
        super(TrimImage, self).__init__(
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype
        )

    def run(self, img):
        imgid = self.get_imgid(img)
        _logger.debug('trimming image %s', imgid)

        img[0] = trimOut(img[0], self.detconf)
        hdr = img['primary'].header
        hdr['NUM-TRIM'] = self.calibid
        hdr['history'] = 'Trimming correction with {}'.format(self.calibid)
        hdr['history'] = 'Trimming correction time {}'.format(datetime.datetime.utcnow().isoformat())

        return img


class GainCorrector(Corrector):
    """A Corrector Node that corrects MEGARA images from different gain."""

    def __init__(self, detconf, datamodel=None, calibid='calibid-unknown', dtype='float32'):
        self.detconf = detconf
        self.gain1 = self.detconf['gain1']
        self.gain2 = self.detconf['gain2']
        super(GainCorrector, self).__init__(
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype
        )

    def run(self, img):
        imgid = self.get_imgid(img)
        _logger.debug('gain correction in image %s', imgid)

        # img[0] = trimOut(img[0], self.detconf)
        hdr = img['primary'].header
        hdr['NUM-GAIN'] = self.calibid
        hdr['BUNIT'] = 'ELECTRON'

        part = img[0].data.shape[0] // 2

        img[0].data[:part] *= self.gain1
        img[0].data[part:] *= self.gain2

        hdr['history'] = 'Gain correction with {}'.format(self.calibid)
        hdr['history'] = 'Gain correction time {}'.format(datetime.datetime.utcnow().isoformat())
        hdr['history'] = 'Gain1 correction value {}'.format(self.gain1)
        hdr['history'] = 'Gain2 correction value {}'.format(self.gain2)
        return img



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline"""


from numina.core import DataFrameType
from numina.types.product import DataProductMixin
from numina.types.datatype import DataType
from numina.types.array import ArrayType
from numina.types.linescatalog import LinesCatalog


from megaradrp.datamodel import MegaraDataModel, QueryAttribute


class MegaraFrame(DataFrameType):
    """A processed frame"""

    tags_headers = {}

    def __init__(self, *args, **kwds):
        super(MegaraFrame, self).__init__(datamodel=MegaraDataModel)


class ProcessedFrame(MegaraFrame):
    """A processed frame"""

    tags_headers = {}


class ProcessedImage(ProcessedFrame):
    """A processed image"""
    pass


class ProcessedRSS(ProcessedFrame):
    """A processed RSS image"""
    pass


class ProcessedMultiRSS(ProcessedFrame):
    """A processed RSS image not to be stored"""
    pass


class ProcessedSpectrum(ProcessedFrame):
    """A 1d spectrum"""
    pass


class ProcessedImageProduct(DataProductMixin, ProcessedImage):
    pass


class ProcessedRSSProduct(DataProductMixin, ProcessedRSS):
    pass


class ProcessedSpectrumProduct(DataProductMixin, ProcessedSpectrum):
    pass


class MegaraLinesCatalog(LinesCatalog):
    __tags__ = {'speclamp': QueryAttribute('speclamp', str), 'vph': QueryAttribute('vph', str)}
    # We are not passing the table of query_attrs in datamodel

    def name(self):
        return "LinesCatalog"


class MasterBias(ProcessedImageProduct):
    """A Master Bias image"""
    pass


class MasterTwilightFlat(ProcessedRSSProduct):
    __tags__ = ['insmode', 'vph', 'confid']


class MasterDark(ProcessedImageProduct):
    """A Master Dark image"""
    pass


class MasterFiberFlat(ProcessedRSSProduct):
    __tags__ = ['insmode', 'vph', 'confid']


class MasterSlitFlat(ProcessedImageProduct):
    __tags__ = ['insmode', 'vph']


class MasterBPM(ProcessedImageProduct):
    """Bad Pixel Mask product"""
    pass


class MasterSensitivity(ProcessedSpectrumProduct):
    """Sensitivity correction."""
    pass


class ReferenceExtinctionTable(DataProductMixin, ArrayType):
    """Atmospheric Extinction."""
    def convert(self, obj):
        # Support None value
        if obj is None:
            return None
        else:
            return super(ReferenceExtinctionTable, self).convert(obj)


class ReferenceSpectrumTable(DataProductMixin, ArrayType):
    """The spectrum of a reference star"""
    pass


class JSONstorage(DataType):
    def __init__(self, default=None):
        super(JSONstorage, self).__init__(ptype=dict, default=default)

    def _datatype_dump(self, obj, where):
        import json
        filename = where + '.json'

        with open(filename, 'w') as fd:
            fd.write(json.dumps(obj, sort_keys=True, indent=2,
                                separators=(',', ': ')))

        return filename

    def _datatype_load(self, obj):
        import json
        try:
            with open(obj, 'r') as fd:
                data = json.load(fd)
        except IOError as e:
            raise e
        return data


class FocusWavelength(JSONstorage):
    """Rich table with focus and wavelength"""
    pass



#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the bpm mode recipe module."""

from numina.core import DataFrame, ObservationResult
import astropy.io.fits as fits
from numina.core.requirements import ObservationResultRequirement
import numina.instrument.assembly as asb

from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.requirements import MasterBiasRequirement, MasterBPMRequirement

from numina.array import combine

from megaradrp.processing.combine import basic_processing_with_combination

from megaradrp.recipes.calibration.tests.test_bpm_common import crear_archivos



class DerivedRecipe(MegaraBaseRecipe):
    obresult = ObservationResultRequirement()
    master_bias = MasterBiasRequirement()
    master_bpm = MasterBPMRequirement()

    def __init__(self, directorio):
        self.directorio = directorio
        super(DerivedRecipe, self).__init__()

    def run(self, rinput):
        import copy

        N = len(rinput.obresult.frames)
        obresult1 = copy.copy(rinput.obresult)
        obresult1.frames = rinput.obresult.frames[:N]

        flow1 = self.init_filters(rinput, rinput.obresult.configuration)
        img = basic_processing_with_combination(rinput, flow1, method=combine.median)
        hdr = img[0].header
        self.set_base_headers(hdr)

        reduced1 = img

        fits.writeto(self.directorio + '/reduced_flat.fits', reduced1[0].data,
                     overwrite=True)

        fits.writeto(self.directorio + '/reduced_flat_bpm.fits',
                     reduced1[0].data, overwrite=True)

        return self.create_result()


def test_bpm_corrector():
    import shutil
    from tempfile import mkdtemp

    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    date_obs = '2017-11-09T11:00:00.0'
    directorio = mkdtemp()
    names = crear_archivos(directorio, number=4)

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'MegaraBiasImage'
    pkg_paths = ['megaradrp.instrument.configs']
    store = asb.load_paths_store(pkg_paths)
    insmodel = asb.assembly_instrument(store, config_uuid, date_obs, by_key='uuid')
    header = fits.Header()
    header['DATE-OBS'] = date_obs
    header['INSCONF'] = config_uuid
    header['INSTRUME'] = 'MEGARA'
    header['VPH'] = 'LR-U'
    header['INSMODE'] = 'MOS'
    insmodel.configure_with_header(header)
    ob.configuration = insmodel
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    recipe = DerivedRecipe(directorio)
    ri = recipe.create_input(obresult=ob, master_bias=DataFrame(
        filename=open(directorio + '/master_bias_data0.fits').name),
                             master_bpm=DataFrame(filename=open(
                                 directorio + '/master_bpm.fits').name))

    recipe.run(ri)
    shutil.rmtree(directorio)

if __name__ == "__main__":
    test_bpm_corrector()




import pytest

from numina.core import BaseRecipe
import numina.core.tagexpr as tagexpr
import numina.types.multitype as mt

from ..loader import load_drp


@pytest.fixture
def current_drp():
    return load_drp()


def simple_tagger(depos, keys):
    result = {}
    for k in keys:
        result[k] = depos[k]
    return result


def test_recipes_are_defined(current_drp):

    assert 'default' in current_drp.pipelines

    for pipeval in current_drp.pipelines.values():
        for key, val in pipeval.recipes.items():
            recipe = pipeval.get_recipe_object(key)
            assert isinstance(recipe, BaseRecipe)

expected_tags = {
    "MegaraSuccess": set(), "MegaraFail": set(), "MegaraBadPixelMask": set(),
    "MegaraBiasImage": set(), "MegaraDarkImage": set(),
    "MegaraArcCalibration": {'insmode', 'speclamp', 'vph'},
    "MegaraSlitFlat": set(), "MegaraTraceMap": set(),
    "MegaraModelMap": {'insmode', 'vph'},
    "MegaraFiberFlatImage": {'insmode', 'vph'},
    "MegaraTwilightFlatImage": {'confid', 'insmode', 'vph'},
    "MegaraFocusSpectrograph": {'insmode', 'vph'},
    "MegaraFocusTelescope": {'confid', 'insmode', 'vph'},
    "MegaraLcbAcquisition": {'confid', 'insmode', 'vph'},
    "MegaraLcbImage": {'confid', 'insmode', 'vph'},
    "MegaraLcbStdStar": {'confid', 'insmode', 'vph'},
    "MegaraMosAcquisition": {'confid', 'insmode', 'vph'},
    "MegaraMosImage": {'confid', 'insmode', 'vph'},
    "MegaraMosStdStar": {'confid', 'insmode', 'vph'},
    "MegaraExtinctionStar": set(), "MegaraSensitivityStar": set()
}


def test_recipes_have_tags(current_drp):

    for pipeval in current_drp.pipelines.values():
        for key, val in pipeval.recipes.items():
            recipe = pipeval.get_recipe_object(key)
            qfields = recipe.tag_names()
            assert qfields == expected_tags[key]


results1 = {
    'MegaraSuccess': {},
    'MegaraFail': {},
    'MegaraBadPixelMask': {'master_bias': 105},
    'MegaraBiasImage': {'master_bpm': 205},
    'MegaraDarkImage': {'master_bias': 105},
    'MegaraArcCalibration': {'master_bias': 105, 'master_bpm': 205, 'master_apertures': 11, 'lines_catalog': 1027},
    'MegaraSlitFlat': {'master_bpm': 205, 'master_bias': 105},
    'MegaraTraceMap': {'master_bias': 105, 'master_bpm': 205},
    'MegaraModelMap': {'master_bpm': 205, 'master_bias': 105, 'master_slitflat': 1, 'master_traces': 11},
    'MegaraFiberFlatImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_traces': 11, 'master_wlcalib': 21},
    'MegaraTwilightFlatImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_traces': 11, 'master_wlcalib': 21, 'master_fiberflat': 49},
    'MegaraFocusSpectrograph': {'master_bias': 105, 'master_bpm': 205, 'master_traces': 11, 'master_wlcalib': 21},
    'MegaraFocusTelescope': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_wlcalib': 21, 'master_fiberflat': 49, 'master_twilight': 59, 'master_traces': 11},
    'MegaraLcbAcquisition': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_wlcalib': 21, 'master_fiberflat': 49, 'master_twilight': 59, 'master_traces': 11},
    'MegaraLcbImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_wlcalib': 21, 'master_fiberflat': 49, 'master_twilight': 59, 'master_traces': 11},
    'MegaraLcbStdStar': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_wlcalib': 21, 'master_fiberflat': 49, 'master_twilight': 59, 'master_traces': 11},
    'MegaraMosAcquisition': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_wlcalib': 21, 'master_fiberflat': 49, 'master_twilight': 59, 'master_traces': 11},
    'MegaraMosImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_wlcalib': 21, 'master_fiberflat': 49, 'master_twilight': 59, 'master_traces': 11},
    'MegaraMosStdStar': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 1, 'master_wlcalib': 21, 'master_fiberflat': 49, 'master_twilight': 59, 'master_traces': 11},
    'MegaraExtinctionStar': {},
    'MegaraSensitivityStar': {},
}


results2 = {
    'MegaraSuccess': {},
    'MegaraFail': {},
    'MegaraBadPixelMask': {'master_bias': 105},
    'MegaraBiasImage': {'master_bpm': 205},
    'MegaraDarkImage': {'master_bias': 105},
    'MegaraArcCalibration': {'master_bias': 105, 'master_bpm': 205, 'master_apertures': 12, 'lines_catalog': 1027},
    'MegaraSlitFlat': {'master_bpm': 205, 'master_bias': 105},
    'MegaraTraceMap': {'master_bias': 105, 'master_bpm': 205},
    'MegaraModelMap': {'master_bpm': 205, 'master_bias': 105, 'master_slitflat': 2, 'master_traces': 12},
    'MegaraFiberFlatImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_traces': 12, 'master_wlcalib': 22},
    'MegaraTwilightFlatImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_traces': 12, 'master_wlcalib': 22, 'master_fiberflat': 42},
    'MegaraFocusSpectrograph': {'master_bias': 105, 'master_bpm': 205, 'master_traces': 12, 'master_wlcalib': 22},
    'MegaraFocusTelescope': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_wlcalib': 22, 'master_fiberflat': 42, 'master_twilight': 52, 'master_traces': 12},
    'MegaraLcbAcquisition': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_wlcalib': 22, 'master_fiberflat': 42, 'master_twilight': 52, 'master_traces': 12},
    'MegaraLcbImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_wlcalib': 22, 'master_fiberflat': 42, 'master_twilight': 52, 'master_traces': 12},
    'MegaraLcbStdStar': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_wlcalib': 22, 'master_fiberflat': 42, 'master_twilight': 52, 'master_traces': 12},
    'MegaraMosAcquisition': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_wlcalib': 22, 'master_fiberflat': 42, 'master_twilight': 52, 'master_traces': 12},
    'MegaraMosImage': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_wlcalib': 22, 'master_fiberflat': 42, 'master_twilight': 52, 'master_traces': 12},
    'MegaraMosStdStar': {'master_bias': 105, 'master_bpm': 205, 'master_slitflat': 2, 'master_wlcalib': 22, 'master_fiberflat': 42, 'master_twilight': 52, 'master_traces': 12},
    'MegaraExtinctionStar': {},
    'MegaraSensitivityStar': {},
}


ob_repo1 = {'vph': 'HR-I', 'insmode': 'MOS', 'confid': '123', 'speclamp': 'ThNe'}
ob_repo2 = {'vph': 'HR-I', 'insmode': 'LCB', 'confid': '000', 'speclamp': 'ThNe'}


@pytest.mark.parametrize("ob_repo, results", [
    (ob_repo1, results1), (ob_repo2, results2)
])
def test_recipes_extract_tags(current_drp, ob_repo, results):

    calibs = {
        'lines_catalog': [
            {'id': 1025, 'tags': {'vph': 'LR-I', 'speclamp': 'ThNe'}},
            {'id': 1026, 'tags': {'vph': 'LR-I', 'speclamp': 'ThAr'}},
            {'id': 1027, 'tags': {'vph': 'HR-I', 'speclamp': 'ThNe'}},
            {'id': 1027, 'tags': {'vph': 'HR-I', 'speclamp': 'ThAr'}},
        ],
        'master_bias': [
            {'id': 105, 'tags': {}},
            {'id': 101, 'tags': {}},
            {'id': 102, 'tags': {}}
        ],
        'master_dark': [
        ],
        'master_bpm': [
            {'id': 205, 'tags': {}},
            {'id': 201, 'tags': {}},
            {'id': 202, 'tags': {}}
        ],
        'master_slitflat': [
            {'id': 5, 'tags': {'vph': 'LR-I', 'insmode': 'MOS'}},
            {'id': 1, 'tags': {'vph': 'HR-I', 'insmode': 'MOS'}},
            {'id': 2, 'tags': {'vph': 'HR-I', 'insmode': 'LCB'}}
        ],
        'master_traces': [
            {'id': 15, 'tags': {'vph': 'LR-I', 'insmode': 'MOS'}},
            {'id': 11, 'tags': {'vph': 'HR-I', 'insmode': 'MOS'}},
            {'id': 12, 'tags': {'vph': 'HR-I', 'insmode': 'LCB'}}
        ],
        'master_wlcalib': [
            {'id': 25, 'tags': {'vph': 'LR-I', 'insmode': 'MOS'}},
            {'id': 21, 'tags': {'vph': 'HR-I', 'insmode': 'MOS'}},
            {'id': 22, 'tags': {'vph': 'HR-I', 'insmode': 'LCB'}}
        ],
        'master_fiberflat': [
            {'id': 45, 'tags': {'vph': 'LR-I', 'insmode': 'MOS', 'confid': '123'}},
            {'id': 49, 'tags': {'vph': 'HR-I', 'insmode': 'MOS', 'confid': '123'}},
            {'id': 41, 'tags': {'vph': 'HR-I', 'insmode': 'MOS', 'confid': '321'}},
            {'id': 42, 'tags': {'vph': 'HR-I', 'insmode': 'LCB', 'confid': '000'}}
        ],
        'master_twilight': [
            {'id': 55, 'tags': {'vph': 'LR-I', 'insmode': 'MOS', 'confid': '123'}},
            {'id': 59, 'tags': {'vph': 'HR-I', 'insmode': 'MOS', 'confid': '123'}},
            {'id': 51, 'tags': {'vph': 'HR-I', 'insmode': 'MOS', 'confid': '321'}},
            {'id': 52, 'tags': {'vph': 'HR-I', 'insmode': 'LCB', 'confid': '000'}}
        ]
    }

    calibs['master_apertures']  = calibs['master_traces']

    for pipeval in current_drp.pipelines.values():
        for key, val in pipeval.recipes.items():
            recipe = pipeval.get_recipe_object(key)

            qfields = recipe.tag_names()

            ob_tags = simple_tagger(ob_repo, qfields)
            match_per_recipe = {}
            for rkey, val in recipe.requirements().items():

                if val.type.isproduct():
                    if isinstance(val.type, mt.MultiType):
                        try_these = val.type.node_type
                    else:
                        try_these = [val.type]

                    for vtype in try_these:
                        expr2 = vtype.query_expr.fill_placeholders(**ob_tags)
                        # Load table of products
                        try:
                            table = calibs[rkey]
                            for entry in table:
                                match = expr2.eval(**entry['tags'])
                                if isinstance(match, tagexpr.Expression):
                                    msg = '{}  tags are not complete: {}'.format(rkey, match)
                                    raise ValueError(msg)

                                if match == True:
                                    match_per_recipe[rkey] = entry['id']
                                    break
                            else:
                                pass
                        except KeyError:
                            pass

            assert results[key] == match_per_recipe



#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import shutil
from tempfile import mkdtemp

import astropy.io.fits as fits
import numpy
from numina.core import DataFrame, ObservationResult
import numina.instrument.assembly as asb

from megaradrp.recipes.calibration.dark import DarkRecipe
from megaradrp.simulation.factory import MegaraImageFactory
from megaradrp.instrument.components.detector import ReadParams, MegaraDetectorSat
from megaradrp.simulation.actions import simulate_dark_fits


def test_dark():

    numpy.random.seed(422992983)

    PSCAN = 50
    DSHAPE = (2056 * 2, 2048 * 2)
    OSCAN = 50

    # ron = 2.0
    ron = 0.001
    gain = 1.0
    bias = 1000.0
    dark = 3.0 # In 1 hour
    exptime = 3600.0
    dark_s = dark / exptime
    qe = 0.8 * numpy.ones(DSHAPE)
    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    date_obs = '2017-11-09T11:00:00.0'
    temporary_path = mkdtemp()
    fits.writeto('{}/eq.fits'.format(temporary_path), qe, overwrite=True)

    readpars1 = ReadParams(gain=gain, ron=ron, bias=bias)
    readpars2 = ReadParams(gain=gain, ron=ron, bias=bias)

    detector = MegaraDetectorSat('megara_test_detector', DSHAPE, OSCAN, PSCAN, qe=qe,
                                 dark=dark_s,
                                 readpars1=readpars1, readpars2=readpars2,
                                 bins='11')

    number = 3
    factory = MegaraImageFactory()
    fs = simulate_dark_fits(factory, detector, exposure=3600, repeat=number)

    for idx, aux in enumerate(fs):
        aux.writeto('{}/dark_{}.fits'.format(temporary_path, idx), overwrite=True)

    header = fits.Header()
    header['DATE-OBS'] = date_obs
    header['INSCONF'] = config_uuid
    header['INSTRUME'] = 'MEGARA'
    header['VPH'] = 'LR-U'
    header['INSMODE'] = 'MOS'
    master_bias_data = numpy.zeros(DSHAPE)
    master_bias_hdul = fits.HDUList(fits.PrimaryHDU(
        master_bias_data, header=header)
    )
    #master_bias_data = master_bias.master_bias.frame[0].data

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'MegaraDarkImage'

    pkg_paths = ['megaradrp.instrument.configs']
    store = asb.load_paths_store(pkg_paths)
    insmodel = asb.assembly_instrument(store, config_uuid, date_obs, by_key='uuid')
    insmodel.configure_with_header(header)
    ob.configuration = insmodel

    names = []
    for aux in range(number):
        names.append('{}/dark_{}.fits'.format(temporary_path, aux))
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    ob.configuration = insmodel
    recipe = DarkRecipe()
    ri = recipe.create_input(
        obresult=ob,
        master_bias=DataFrame(frame=master_bias_hdul),
    )
    aux = recipe.run(ri)

    mean_dark_value = aux.master_dark.frame[0].data.mean()

    shutil.rmtree(temporary_path)

    assert numpy.allclose(mean_dark_value, dark, atol=0, rtol=1e-1)


if __name__ == "__main__":
    test_dark()



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Acquisition with the Fiber MOS"""


import math

import numpy as np
from numina.array.offrot import fit_offset_and_rotation
from numina.core import Result, Parameter
from numina.core.qc import QC

from megaradrp.datamodel import TargetType
from megaradrp.recipes.scientific.base import ImageRecipe
from megaradrp.types import ProcessedRSS, ProcessedFrame
from megaradrp.utils import add_collapsed_mos_extension


class AcquireMOSRecipe(ImageRecipe):
    """Process Acquisition MOS images.

    This recipe processes a set of acquisition images
    obtained in **MOS Acquisition** mode and returns
    the offset and rotation required to center the
    fiducial objects in their reference positions.

    See Also
    --------
    megaradrp.recipes.auxiliary.acquisitionlcb.AcquireLCBRecipe
    numina.array.offrot: Kabsch algorithm for offset and rotation

    Notes
    -----
    Images provided by `obresult` are trimmed and corrected
    from overscan, bad pixel mask (if `master_bpm` is not None),
    bias, dark current (if `master_dark` is not None) and
    slit-flat (if `master_slitflat` is not None).

    Images thus corrected are the stacked using the median.
    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces` and resampled according to the wavelength calibration in
    `master_wlcalib`. Then is divided by the `master_fiberflat`.
    The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    The sky is subtracted by combining the the fibers marked as `SKY`
    in the fibers configuration. The RSS with sky subtracted is returned ini the
    field `final_rss` of the recipe result.

    Then, the centroid of each fiducial object, marked as `REFERENCE` in the fibers
    configuration, is computed. The offset and rotation needed to center
    each fiducial object in its bundle is computed and returned

    """

    # Requirements are defined in base class
    extraction_region = Parameter(
        [1000, 3000],
        description='Region used to compute a mean flux',
        nelem=2
    )

    reduced_image = Result(ProcessedFrame)
    reduced_rss = Result(ProcessedRSS)
    final_rss = Result(ProcessedRSS)
    offset = Result(list)
    rotang = Result(float)

    def run(self, rinput):

        self.logger.info('starting AC MOS reduction')

        reduced2d, reduced1d = super(AcquireMOSRecipe, self).base_run(rinput)

        do_sky_subtraction = True
        if do_sky_subtraction:
            self.logger.info('start sky subtraction')
            isb = rinput.ignored_sky_bundles
            if isb:
                self.logger.info('sky bundles ignored: %s', isb)
            final, origin, sky = self.run_sky_subtraction(reduced1d,
                                                          ignored_sky_bundles=isb)
            self.logger.info('end sky subtraction')
        else:
            final =  reduced1d
            origin = final
            sky = final

        fiberconf = self.datamodel.get_fiberconf(final)

        cut1, cut2 = rinput.extraction_region

        self.logger.debug("MOS configuration is %s", fiberconf.conf_id)
        rssdata = final[0].data
        scale, funit = self.datamodel.fiber_scale_unit(final, unit=True)
        self.logger.debug('unit is %s', funit)
        platescale = self.datamodel.PLATESCALE

        p1 = []
        q1 = []
        temp = []
        for key, bundle in fiberconf.bundles.items():
            if bundle.target_type == TargetType.REFERENCE:
                self.logger.debug("%s %s %s", key, bundle.target_name, bundle.target_type)
                sorted_fibers = [bundle.fibers[key] for key in sorted(bundle.fibers)]
                central_fiber = sorted_fibers[3] # Central fiber is number 4 in the list
                central_coords = [central_fiber.x * scale, central_fiber.y * scale]
                #central_fiber_pair_id
                # Central fiber is
                self.logger.debug('Center fiber is %d', central_fiber.fibid)
                self.logger.debug('Center fiber coordinates %f %f arcsec', 
                             central_fiber.x * scale, central_fiber.y * scale)

                colids = []
                coords = []
                for fiber in sorted_fibers:
                    colids.append(fiber.fibid - 1)
                    coords.append((fiber.x, fiber.y))
                self.logger.debug('nearest fibers')
                self.logger.debug('%s', [col + 1 for col in colids])
                coords = np.asarray(coords) * scale
                flux_per_cell = rssdata[colids, cut1:cut2].mean(axis=1)
                flux_per_cell_total = flux_per_cell.sum()
                flux_per_cell_norm = flux_per_cell / flux_per_cell_total
                # centroid
                scf = coords.T * flux_per_cell_norm
                centroid = scf.sum(axis=1)
                self.logger.info('centroid: %s arcsec', list(centroid))
                self.logger.info('centroid: %s mm', list(centroid / platescale))
                # central coords
                c_coords = coords - centroid
                scf0 = scf - centroid[:, np.newaxis] * flux_per_cell_norm
                mc2 = np.dot(scf0, c_coords)
                self.logger.info('2nd order moments, x2=%f, y2=%f, xy=%f arcsec^2', mc2[0, 0], mc2[1, 1], mc2[0, 1])

                p1.append(central_coords)
                q1.append(centroid)
                temp.append((bundle.id, central_fiber.fibid, central_fiber.x * scale, central_fiber.y * scale, centroid[0], centroid[1]))

        if self.intermediate_results:
            with open("centroids.txt", "w") as fd:
                for entry in temp:
                    fd.write("%s %s %6.3f %6.3f %6.3f %6.3f\n" % entry)

        self.logger.info('compute offset and rotation with %d points', len(p1))
        if len(p1) == 0:
            self.logger.warn('cant compute offset and rotation with 0 points')
            offset = [0.0, 0.0]
            angle = 0.0
            qc = QC.BAD
        else:
            offset, rot = fit_offset_and_rotation(np.array(p1), np.array(q1))
            angle = math.atan2(rot[1, 0], rot[0, 0])
            angle = angle / math.pi * 180.0
            qc = QC.GOOD
            self.logger.info('offset is %s', offset)
            self.logger.info('rot matrix is %s', rot)
            self.logger.info('rot angle %5.2f deg', angle)

        final = add_collapsed_mos_extension(final)
        origin = add_collapsed_mos_extension(origin)

        return self.create_result(
            reduced_image=reduced2d,
            reduced_rss=reduced1d,
            final_rss=final,
            offset=offset,
            rotang=angle,
            qc=qc
        )




import numpy

import megaradrp.processing.extractobj as eobj


def test_cover_1():

    arr = [1, 1, 1, 1, 1, 1]
    res = eobj.coverage_det(arr)
    assert res == slice(0, 7)

    arr = [0, 0, 0, 0, 0, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(0, 0)


def test_cover_2():

    arr = [0, 1, 1, 1, 0, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(1, 4, None)

    arr = [0, 1, 1, 1, 1, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(1, 5)

    arr = [0, 1, 1, 1, 1, 1]
    res = eobj.coverage_det(arr)
    assert res == slice(1, 7)

    arr = [1, 1, 1, 1, 1, 0]
    res = eobj.coverage_det(arr)
    assert res == slice(0, 5)


def test_coverage_issue247():

    # Example from test case
    fmap = numpy.zeros((4300,), dtype='int32')
    fmap[226:4136] = 37
    fmap[224] = 1
    fmap[225] = 10
    fmap[4136] = 32

    # Test total coverage
    arr1 = fmap == fmap.max()
    res1 = eobj.coverage_det(arr1.astype('int'))
    assert res1 == slice(226, 4136, None)

    # Test max coverage
    arr2 = fmap > 0
    res2 = eobj.coverage_det(arr2.astype('int'))
    assert res2 == slice(224, 4137, None)



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy as np
import math

from astropy.modeling.functional_models import Fittable2DModel, Parameter


_HEX_SCALE = 0.25 * math.sqrt(3.0)


class HexagonA(Fittable2DModel):

    amplitude = Parameter(default=1)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)
    radius = Parameter(default=1)
    angle = Parameter(default=0)

    @staticmethod
    def evaluate(x, y, amplitude, x_0, y_0, radius, angle):
        # Recenter
        xo = x - x_0
        yo = y - y_0
        # Rotate
        cs = math.cos(angle)
        ss = math.sin(angle)
        xr = cs * xo + ss * yo
        yr = -ss * xo + cs * yo
        # Scale
        d = radius * 2
        dx = np.abs(xr) / d
        dy = np.abs(yr) / d
        a = 0.25 * math.sqrt(3.0)
        sel1 = (dy <= a)
        sel2 = (a * dx + 0.25 * dy <= 0.5 * a)
        return np.select([np.logical_and(sel1, sel2)],
                         [amplitude], 0)

    @property
    def bounding_box(self):
        a = 0.25 * math.sqrt(3.0)
        d = 2 * self.radius
        return ((self.y_0 - a * d, self.y_0 + a * d),
                (self.x_0 - self.radius, self.x_0 + self.radius))


# Returns True in inside hexagon, rectangle and square
def hex_c(x, y, rad, ang=0.0):
    # Rotate coordinates
    cs = math.cos(ang)
    ss = math.sin(ang)
    xr = cs * x + ss *y
    yr = -ss * x + cs *y
    d = rad * 2
    dx = np.abs(xr) / d
    dy = np.abs(yr) / d
    a = 0.25 * math.sqrt(3.0)
    return (dy <= a) & (a*dx + 0.25*dy <= 0.5*a)


def rect_c(x, y, lx, ly):
    dx = np.abs(x) / lx
    dy = np.abs(y) / ly
    return (dy <= 0.5) & (dx <= 0.5)


def square_c(x, y, l):
    return rect_c(x, y, l, l)


def setup_grid(xsize, ysize, Dx, Dy):

    # Simulation size
    # xsize = 12.5
    # ysize = 12.5
    # Pixel size
    # Dx = 0.005
    # Dy = 0.005

    # Borders
    xl = xsize / 2.0
    yl = ysize / 2.0

    xs = np.arange(-xl, xl + Dx, Dx)
    ys = np.arange(-yl, yl + Dy, Dy)

    # print 'simulated images are', nxpoints, 'x', nypoints

    # grid
    xx, yy = np.meshgrid(xs, ys, sparse=False, indexing='xy')

    return xx, yy, xs, ys, xl, yl



#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging
import datetime

from .fiberflat import CommonFlatCorrector

_logger = logging.getLogger(__name__)


class SlitFlatCorrector(CommonFlatCorrector):
    """A Node that corrects a frame from slit flat."""

    def __init__(self, slitflat, datamodel=None, calibid='calibid-unknown',
                 dtype='float32'):

        super(SlitFlatCorrector, self).__init__(slitflat, datamodel, calibid, dtype)
        self.flattag = 'slitflat'

    def header_update(self, hdr, imgid):
        hdr['NUM-SLTF'] = self.calibid
        hdr['history'] = 'Slit flat correction {}'.format(imgid)
        hdr['history'] = 'Slit flat correction time {}'.format(datetime.datetime.utcnow().isoformat())



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging

from numina.instrument.simulation.factory import PersistentRunCounter

from megaradrp.simulation.actions import megara_sequences
from megaradrp.simulation.efficiency import InterpolFile


_logger = logging.getLogger("simulation")


class PointLike(object):
    def __init__(self, factor=1.0, **kwds):
        self.factor = factor


class Target(object):
    def __init__(self, name, relposition, profile, spectrum):
        self.name = name
        self.relposition = relposition
        self.profile = profile
        self.spectrum = spectrum


def profile_builder(profile_entry):
    profile_type = profile_entry['type']
    profile_args = profile_entry.get('args', {})
    if profile_type == 'point-like':
        return PointLike(**profile_args)
    else:
        raise ValueError("'{0}' not registered", profile_type)


class ControlSystem(object):
    """Top level"""
    def __init__(self, factory):
        self._elements = {}

        self.imagecount = PersistentRunCounter('r00%04d.fits')
        self.mode = 'null'
        self.ins = 'MEGARA'
        self.seqs = megara_sequences()
        self.factory = factory
        self.ob_data = dict(count=0, repeat=0, name=None, obsid=1)
        self.targets = None

    def register(self, name, element):
        self._elements[name] = element

    def get(self, name):
        return self._elements[name]

    def set_mode(self, mode):
        self.mode = mode

    def set_targets(self, targets_description):
        # Process targets here

        targets = targets_description['targets']

        tlist = []

        for m in targets:
            entry = targets[m]
            profile_entry = entry['profile']
            profile = profile_builder(profile_entry)
            spectrum_entry = entry['spectrum']
            filename = spectrum_entry['sed']
            factor = spectrum_entry.get('factor', 1.0)
            spectrum_entry['sed'] = InterpolFile(filename, factor=factor)
            spectrum = spectrum_entry
            relposition = entry['relposition']
            tar = Target(m, relposition, profile=profile, spectrum=spectrum)
            tlist.append(tar)

        self.targets = tlist

    def run(self, exposure, repeat=1):

        if repeat < 1:
            return

        _logger.info('mode is %s', self.mode)
        try:
            thiss = self.seqs[self.mode]
        except KeyError:
            _logger.error('No sequence for mode %s', self.mode)
            raise

        iterf = thiss.run(self, exposure, repeat)

        self.ob_data['repeat'] = repeat
        self.ob_data['name'] = None
        for count, final in enumerate(iterf, 1):
            _logger.info('image %d of %d', count, repeat)
            self.ob_data['name'] = self.imagecount.runstring()
            self.ob_data['count'] = count
            fitsfile = self.factory.create(final, self.ob_data['name'], self)
            _logger.info('save image %s', self.ob_data['name'])
            fitsfile.writeto(self.ob_data['name'], overwrite=True)

    def config_info(self):
        return {'ob_data': self.ob_data}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.imagecount.__exit__(exc_type, exc_val, exc_tb)



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


import astropy.io.fits as fits
import astropy.table
import pytest

from ..datamodel import MegaraDataModel, FibersConf


def create_empty_img(insmode):
    img = fits.HDUList([fits.PrimaryHDU()])
    img[0].header['insmode'] = insmode
    return img


BASE_LCB = ("LCB", 'b7dcd9d1-0b60-4b43-b26e-d2c9868d5e20', 9, 623)
BASE_MOS = ("MOS", '00000000-0000-0000-0000-000000000000', 92, 644)
FIBS_SKY_ALL = [22, 23, 24, 25, 26, 27, 28,
                57, 58, 59, 60, 61, 62, 63,
                134, 135, 136, 137, 138, 139, 140,
                267, 268, 269, 270, 271, 272, 273,
                351, 352, 353, 354, 355, 356, 357,
                484, 485, 486, 487, 488, 489, 490,
                561, 562, 563, 564, 565, 566, 567,
                596, 597, 598, 599, 600, 601, 602
                ]
FIBS_SKY_94_99 = [57, 58, 59, 60, 61, 62, 63,
                  561, 562, 563, 564, 565, 566, 567,
                  ]

@pytest.mark.parametrize("name, confid, nbundles, nfibers",
                         [BASE_LCB, BASE_MOS])
def test_fiberconf_1(name, confid, nbundles, nfibers):

    datamodel = MegaraDataModel()

    img = create_empty_img(name)

    conf = datamodel.get_fiberconf(img)

    assert isinstance(conf, FibersConf)
    # Default values from file
    assert conf.name == name
    assert conf.conf_id == confid
    assert conf.nbundles == nbundles
    assert conf.nfibers ==  nfibers


def test_fiberconf_other():

    datamodel = MegaraDataModel()

    img = create_empty_img('OTHER')

    with pytest.raises(ValueError):
        datamodel.get_fiberconf(img)


@pytest.mark.parametrize("name, confid, nbundles, nfibers",
                         [BASE_LCB, BASE_MOS])
def test_bundles_to_table(name, confid, nbundles, nfibers):

    datamodel = MegaraDataModel()

    img = create_empty_img(name)

    conf = datamodel.get_fiberconf(img)

    bundles_t = conf.bundles_to_table()
    assert isinstance(bundles_t, astropy.table.Table)
    assert len(bundles_t) == nbundles
    assert bundles_t.colnames == ['bundle_id', 'x', 'y', 'pa', 'enabled',
                                  'target_type', 'target_priority', 'target_name']


@pytest.mark.parametrize("name, confid, nbundles, nfibers",
                         [BASE_LCB, BASE_MOS])
def test_fibers_to_table(name, confid, nbundles, nfibers):

    datamodel = MegaraDataModel()

    img = create_empty_img(name)

    conf = datamodel.get_fiberconf(img)

    fibers_t = conf.fibers_to_table()
    assert isinstance(fibers_t, astropy.table.Table)
    assert len(fibers_t) == nfibers
    assert fibers_t.colnames == ['fibid', 'name', 'x', 'y',
                                 'inactive', 'valid',
                                 'bundle_id']


@pytest.mark.parametrize("ignored, fibs",
                         [([93,94,95,96,97,98,99,100], []),
                          ([93,95,96,97,98,100], FIBS_SKY_94_99),
                          ([], FIBS_SKY_ALL),
                          (None, FIBS_SKY_ALL),
                          ])
def test_sky_fibers(ignored, fibs):

    datamodel = MegaraDataModel()

    img = create_empty_img('LCB')

    conf = datamodel.get_fiberconf(img)

    skyfibs = conf.sky_fibers(valid_only=True,
                              ignored_bundles=ignored
                              )
    assert sorted(skyfibs) == fibs





#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the bpm mode recipe module."""
import shutil
from tempfile import mkdtemp

import astropy.io.fits as fits
import numpy as np

from numina.core import DataFrame, ObservationResult
import numina.instrument.assembly as asb

from megaradrp.recipes.calibration.bpm import BadPixelsMaskRecipe
from megaradrp.recipes.calibration.tests.test_bpm_common import generate_bias
from megaradrp.instrument.components.detector import ReadParams, MegaraDetectorSat
from megaradrp.simulation.actions import simulate_flat


# @pytest.mark.remote_data
def test_bpm():
    number = 5
    PSCAN = 50
    DSHAPE = (2056 * 2, 2048 * 2)
    OSCAN = 50

    ron = 2.0
    gain = 1.0
    bias = 1000.0

    qe = 0.8 * np.ones(DSHAPE)
    qe[5:6, 0:170] = 0.0
    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    date_obs = '2017-11-09T11:00:00.0'
    temporary_path = mkdtemp()

    fits.writeto('{}/eq.fits'.format(temporary_path), qe, overwrite=True)

    readpars1 = ReadParams(gain=gain, ron=ron, bias=bias)
    readpars2 = ReadParams(gain=gain, ron=ron, bias=bias)

    detector = MegaraDetectorSat('megara_test_detector', DSHAPE, OSCAN, PSCAN,
                                 qe=qe,
                                 dark=(3.0 / 3600.0),
                                 readpars1=readpars1, readpars2=readpars2,
                                 bins='11')

    source2 = 1.0

    fs = [simulate_flat(detector, exposure=1.0, source=5000 * source2) for i in
          range(number)]
    fs2 = [simulate_flat(detector, exposure=1.0, source=40000 * source2) for i
           in range(number)]

    header = fits.Header()
    header['DATE-OBS'] = date_obs
    header['INSCONF'] = config_uuid
    header['INSTRUME'] = 'MEGARA'
    header['VPH'] = 'LR-U'
    header['INSMODE'] = 'MOS'
    for aux in range(len(fs)):
        fits.writeto('{}/flat_{}.fits'.format(temporary_path, aux), fs[aux],
                     header=header,
                     overwrite=True)
        fits.writeto('{}/flat_{}.fits'.format(temporary_path, aux + number), fs2[aux],
                     header=header,
                     overwrite=True)

    result = generate_bias(detector, number, temporary_path)
    result.master_bias.frame.writeto(
        '{}/master_bias_data0.fits'.format(temporary_path),
        overwrite=True
    )

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'MegaraBiasImage'
    pkg_paths = ['megaradrp.instrument.configs']
    store = asb.load_paths_store(pkg_paths)
    insmodel = asb.assembly_instrument(store, config_uuid, date_obs, by_key='uuid')
    insmodel.configure_with_header(header)
    ob.configuration = insmodel

    names = []
    for aux in range(number * 2):
        names.append('{}/flat_{}.fits'.format(temporary_path, aux))
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    recipe = BadPixelsMaskRecipe()
    ri = recipe.create_input(obresult=ob, master_bias=DataFrame(
        filename=open(temporary_path + '/master_bias_data0.fits').name))
    aux = recipe.run(ri)
    aux.master_bpm.frame.writeto('{}/master_bpm.fits'.format(temporary_path), overwrite=True)
    shutil.rmtree(temporary_path)


if __name__ == "__main__":
    test_bpm()



#
# Copyright 2014-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Load MEGARA DRP"""

from numina.core import drp_load


def load_drp():
    """Entry point to load MEGARA DRP."""
    return drp_load('megaradrp', 'drp.yaml')



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Sequences for observing modes of MEGARA"""

import math
import logging

import numpy as np
from scipy.interpolate import RectBivariateSpline
from scipy import signal
import scipy.spatial
from astropy import units as u
import astropy.constants as cons
from numina.instrument.simulation.actions import Sequence

from .convolution import rect_c, hex_c, setup_grid


_logger = logging.getLogger(__name__)


def simulate_bias(detector):
    """Simulate a BIAS array."""
    detector.expose(source=0.0, time=0.0)
    final = detector.readout()
    return final


def simulate_dark(detector, exposure):
    """Simulate a DARK array,"""
    detector.expose(source=0.0, time=exposure)
    final = detector.readout()
    return final


def simulate_flat(detector, exposure, source):
    """Simulate a FLAT array,"""
    detector.expose(source=source, time=exposure)
    final = detector.readout()
    return final


def simulate_dark_fits(factory, instrument, exposure, repeat=1):
    """Simulate a DARK FITS."""
    # Use instrument and detector!
    det = getattr(instrument, 'detector', None)
    if det:
        detector = det
    else:
        detector = instrument

    for i in range(repeat):
        detector.expose(time=exposure)
        final = detector.readout()
        fitsfile = factory.create_from_instrument(mode='MegaraDarkImage',
                                                  instrument=detector,
                                                  data=final,
                                                  name='dark_%s.fits'%i
                                                  )

        yield fitsfile


class MegaraSequence(Sequence):
    def __init__(self, mode):
        super(MegaraSequence, self).__init__('MEGARA', mode)


class MegaraNullSequence(Sequence):
    def __init__(self):
        super(MegaraNullSequence, self).__init__('MEGARA', 'null')

    def run(self, control, exposure, repeat):
        # This is an empty generator
        return iter(())


class MegaraBiasSequence(MegaraSequence):
    def __init__(self):
        super(MegaraBiasSequence, self).__init__('MegaraBiasImage')

    def setup_instrument(self, instrument):
        instrument.shutter = 'STOP'

    def run(self, control, exposure, repeat):
        instrument = control.get(self.instrument)

        self.setup_instrument(instrument)

        for i in range(repeat):
            instrument.detector.expose()
            final = instrument.detector.readout()
            yield final


class MegaraDarkSequence(MegaraSequence):
    def __init__(self):
        super(MegaraDarkSequence, self).__init__('MegaraDarkImage')

    def setup_instrument(self, instrument):
        instrument.shutter = 'STOP'

    def run(self, control, exposure, repeat):
        instrument = control.get(self.instrument)

        self.setup_instrument(instrument)

        for i in range(repeat):
            instrument.detector.expose(source=0.0, time=exposure)
            final = instrument.detector.readout()
            yield final


class MegaraLampSequence(Sequence):
    def __init__(self, mode):
        super(MegaraLampSequence, self).__init__('MEGARA', mode)

    def setup_instrument(self, instrument):
        instrument.shutter = 'OPEN'

    def run(self, control, exposure, repeat):
        instrument = control.get(self.instrument)
        cu = control.get('ICM-MEGARA')

        self.setup_instrument(instrument)

        # Get active lamp
        lamp = cu.current()
        if lamp == 'EMPTY':
            raise ValueError('LAMP is %s, exiting' %  lamp)

        self.lamp_check(lamp)
        # Simulated arc spectrum

        wl_in, lamp_illum = self.lamp_in_focal_plane(lamp, instrument)
        out = instrument.apply_transmissions(wl_in, lamp_illum)

        for i in range(repeat):
            instrument.detector.expose(source=out, time=exposure)
            final = instrument.detector.readout()
            yield final

    def lamp_check(self, lamp):
        raise NotImplemented

    def lamp_in_focal_plane(self, lamp, instrument):

        # print('enter targets in focal plane')
        wl = instrument.vph.wltable_interp()
        # Total number of fibers, defines RSS size
        nfibers = instrument.fiberset.nfibers
        # Radius of the fiber (on the focal plane, arcsecs)
        fibrad = instrument.fiberset.size
        fibarea = instrument.fiberset.area
        # Result, RSS like (with higher resolution in WL)
        final = np.zeros((nfibers, wl.shape[0]))

        # Centers of profiles
        tab = instrument.focal_plane.get_visible_fibers(instrument.fiberset.name)
        fibid = tab['fibid']
        pos_x = tab['x']
        pos_y = tab['y']
        cover_frac = tab['cover']

        # fraction of flux in each fiber
        scales = np.zeros((nfibers,))

        # scales include the cover
        scales[fibid - 1] = cover_frac

        if lamp.illumination:
            scales[fibid - 1] *= lamp.illumination(pos_x, pos_y)

        photon_energy = (cons.h * cons.c / wl)

        sed = lamp.flux(wl)
        area = 1.0 * u.cm**2 # FIXME, area of the lamp?
        nphot_area_t_wl_cgs = (sed / photon_energy * area * fibarea).to(u.s**-1 * u.micron**-1)

        final += scales[:, np.newaxis] * nphot_area_t_wl_cgs[np.newaxis, :]

        return wl, final


class MegaraFiberFlatSequence(MegaraLampSequence):
    def __init__(self):
        super(MegaraFiberFlatSequence, self).__init__('MegaraFiberFlatImage')

    def lamp_check(self, lamp):
        # TODO: check that this is a cont lamp
        return True


class MegaraTraceMapSequence(MegaraLampSequence):
    def __init__(self):
        super(MegaraTraceMapSequence, self).__init__('MegaraTraceAmp')

    def lamp_check(self, lamp):
        # TODO: check that this is a cont lamp
        return True


class MegaraArcSequence(MegaraLampSequence):
    def __init__(self):
        super(MegaraArcSequence, self).__init__('MegaraArcCalibration')

    def lamp_check(self, lamp):
        # TODO: check that this is an arc lamp
        return True


class MegaraSlitFlatSequence(MegaraLampSequence):
    def __init__(self):
        super(MegaraSlitFlatSequence, self).__init__('MegaraSlitFlat')

    def setup_instrument(self, instrument):
        instrument.shutter = 'OPEN'

        # Internal focus, we need to defocus a lot
        instrument.set_focus(1000)
        # FIXME: seting internal value directly
        instrument._internal_focus_factor = 8

    def run(self, control, exposure, repeat):

        instrument = control.get(self.instrument)
        cu = control.get('ICM-MEGARA')

        self.setup_instrument(instrument)

        # Get active lamp
        lamp = cu.current()
        if lamp == 'EMPTY':
            raise ValueError('LAMP is %s, exiting' % lamp)

        # Simulated arc spectrum
        wl_in, lamp_illum = self.lamp_in_focal_plane(lamp, instrument)
        out = instrument.apply_transmissions(wl_in, lamp_illum)

        for i in range(repeat):
            instrument.detector.expose(source=out, time=exposure)
            final = instrument.detector.readout()
            yield final


class MegaraTwilightFlatSequence(Sequence):
    def __init__(self):
        super(MegaraTwilightFlatSequence, self).__init__('MEGARA', 'MegaraTwilightFlatImage')

    def setup_instrument(self, instrument):
        instrument.shutter = 'OPEN'

    def run(self, control, exposure, repeat):
        instrument = control.get(self.instrument)
        telescope = control.get('GTC')
        atm = telescope.inc # Atmosphere model
        oe = control.get('OE')

        self.setup_instrument(instrument)
        # Simulated tw spectrum

        targets2 = [atm.twilight_spectrum]

        wl_in, ns_illum = all_targets_in_focal_plane([], targets2, [], atm, telescope, oe, instrument)
        out1 = instrument.apply_transmissions_only(wl_in, ns_illum)
        out2 = instrument.project_rss(wl_in, out1)

        for i in range(repeat):
            instrument.detector.expose(source=out2, time=exposure)
            final = instrument.detector.readout()
            yield final


class MegaraFocusSequence(MegaraLampSequence):
    def __init__(self):
        super(MegaraFocusSequence, self).__init__(mode='MegaraFocusSpectrograph')

    def run(self, control, exposure, repeat):
        instrument = control.get(self.instrument)
        instrument.shutter = 'OPEN'
        cu = control.get('ICM-MEGARA')

        # Get active lamp
        lamp = cu.current()
        self.lamp_check(lamp)
        # Simulated arc spectrum
        wl_in, lamp_illum = self.lamp_in_focal_plane(lamp, instrument)

        # FIXME, hardcoded
        focii = range(-1000, 4000, 1000)
        for focus in focii:
            instrument.set_focus(focus)
            out = instrument.apply_transmissions(wl_in, lamp_illum)

            for i in range(repeat):
                instrument.detector.expose(source=out, time=exposure)
                final = instrument.detector.readout()
                yield final

    def lamp_check(self, lamp):
        return True


class MegaraSkyImageSequence(MegaraSequence):
    def __init__(self, mode):
        super(MegaraSkyImageSequence, self).__init__(mode)

    def setup_instrument(self, instrument):
        # Setup instrument
        # Fixed by mode
        instrument.shutter = 'OPEN'

    def run(self, control, exposure, repeat):

        instrument = control.get(self.instrument)

        self.setup_instrument(instrument)

        telescope = control.get('GTC')
        oeng = control.get('OE')
        atm = telescope.inc # Atmosphere model, incoming light

        # Get targets
        targets1 = control.targets
        targets2 = [atm.night_spectrum]

        wl_in, ns_illum = all_targets_in_focal_plane(targets1, targets2, [], atm, telescope, oeng, instrument)

        out1 = instrument.apply_transmissions_only(wl_in, ns_illum)
        out2 = instrument.project_rss(wl_in, out1)

        for i in range(repeat):
            instrument.detector.expose(source=out2, time=exposure)
            final = instrument.detector.readout()
            yield final


class MegaraSkyLCBImageSequence(MegaraSkyImageSequence):
    def __init__(self, mode):
        super(MegaraSkyLCBImageSequence, self).__init__(mode)

    def setup_instrument(self, instrument):
        super(MegaraSkyLCBImageSequence, self).setup_instrument(instrument)
        instrument.insmode = 'LCB'


class MegaraLCBImageSequence(MegaraSkyLCBImageSequence):
    def __init__(self):
        super(MegaraLCBImageSequence, self).__init__('MegaraLcbImage')


class MegaraLCBAcquisitionSequence(MegaraSkyLCBImageSequence):
    def __init__(self):
        super(MegaraLCBAcquisitionSequence, self).__init__('MegaraLcbAcquisition')


class MegaraFocusTelescopeSequence(MegaraSkyLCBImageSequence):
    def __init__(self):
        super(MegaraFocusTelescopeSequence, self).__init__('MegaraFocusTelescope')

    def run(self, control, exposure, repeat):
        telescope = control.get('GTC')

        # FIXME, hardcoded
        focii = range(-3000, 4000, 1000)
        for focus in focii:
            telescope.set_focus(focus)

            for img in super(MegaraFocusTelescopeSequence, self).run(control, exposure, repeat):
                yield img


class MegaraSkyMOSImageSequence(MegaraSkyImageSequence):
    def __init__(self, mode):
        super(MegaraSkyMOSImageSequence, self).__init__(mode)

    def setup_instrument(self, instrument):
        super(MegaraSkyMOSImageSequence, self).setup_instrument(instrument)
        instrument.insmode = 'MOS'


class MegaraMOSImageSequence(MegaraSkyMOSImageSequence):
    def __init__(self):
        super(MegaraMOSImageSequence, self).__init__('MegaraMosImage')


class MegaraMOSAcquisitionSequence(MegaraSkyMOSImageSequence):
    def __init__(self):
        super(MegaraMOSAcquisitionSequence, self).__init__('MegaraMosAcquisition')


def megara_sequences():
    seqs = {}
    # Keys must correspod to MEGARA ObsMode.key
    seqs['null'] = MegaraNullSequence()
    seqs['bias_image'] = MegaraBiasSequence()
    seqs['dark_image'] = MegaraDarkSequence()
    seqs['fiber_flat_image'] = MegaraFiberFlatSequence()
    seqs['slit_flat'] = MegaraSlitFlatSequence()
    seqs['trace_map'] = MegaraTraceMapSequence()
    seqs['arc_calibration'] = MegaraArcSequence()
    seqs['twilight_flat_image'] = MegaraTwilightFlatSequence()
    seqs['focus_spectrograph'] = MegaraFocusSequence()
    seqs['focus_telescope'] = MegaraFocusTelescopeSequence()
    seqs['lcb_image'] = MegaraLCBImageSequence()
    seqs['mos_image'] = MegaraMOSImageSequence()
    seqs['mos_acquisition'] = MegaraMOSAcquisitionSequence()
    seqs['lcb_acquisition'] = MegaraLCBAcquisitionSequence()
    return seqs


def simulate_point_like_profile(seeing_profile, psf, fibrad, angle=0.0, xsize=12.5, ysize=12.5):
    # Simulation of the fraction of flux in each spaxel
    # By convolution of the seeing profile with the
    # spaxel shape
    # Simulation size
    # xsize, ysize
    # Pixel size for simulation
    Dx = 0.005
    Dy = 0.005

    xx, yy, xs, ys, xl, yl = setup_grid(xsize, ysize, Dx, Dy)
    # print('generate hex kernel')
    # fibrad without units
    hex_kernel = hex_c(xx, yy, rad=fibrad, ang=angle / 180.0 * math.pi)

    # print('generate seeing profile')
    sc = seeing_profile(xx, yy)

    # print('convolve hex kernel with seeing profile')
    # Convolve model gaussian with hex kernel
    convolved = signal.fftconvolve(hex_kernel, sc, mode='same')

    # print('setup 2D interpolator for point-like object')
    # Interpolate
    rbs = RectBivariateSpline(xs, ys, convolved)

    def fraction_of_flux(offpos0, offpos1):

        # Check those that are inside the convolved image
        result = np.zeros_like(offpos0)
        validfibs = rect_c(offpos0, offpos1, 2 * xl, 2 * yl)

        # import matplotlib.pyplot as plt
        # plt.imshow(sc, extent=[-xl-0.5*Dx, xl+0.5*Dx, -yl-0.5*Dy,yl+0.5*Dy], origin='lower')
        # plt.scatter(offpos0[validfibs], offpos1[validfibs])
        # plt.show()

        # Sample convolved image
        # In the positions of active fibers
        #
        values = rbs.ev(offpos0[validfibs], offpos1[validfibs]) * Dx * Dy
        # scales include the cover
        result[validfibs] = values
        return result

    return fraction_of_flux


def all_targets_in_focal_plane(t1, t2, t3, atmosphere, telescope, oe, instrument):
    # simulate_seeing_profile

    # print('enter targets in focal plane')
    # WL in microns
    wl = instrument.vph.wltable_interp()
    # Total number of fibers, defines RSS size
    nfibers = instrument.fiberset.nfibers
    # Radius of the fiber (on the focal plane, arcsecs)
    fibrad = instrument.fiberset.size
    fibarea = instrument.fiberset.area

    # Result, RSS like (with higher resolution in WL)
    final = np.zeros((nfibers, wl.shape[0]))

    # Centers of profiles
    # zenith_distance = 30 * u.deg
    # airmass = 1.0 / math.cos(zenith_distance.to(u.rad).value)

    if instrument.fiberset.name == 'MOS':
        # get fiber mos
        fibermos = instrument.get_device('MEGARA.MOS')
        ids, pos = fibermos.robots_in_focal_plane()
        pos = np.array(pos)

        maxdis = 20.0
        base = scipy.spatial.KDTree(pos)
        for target in t1:
            _logger.debug('simulate MOS')
            _logger.debug('find robot nearest to %s', target.relposition)
            result = base.query(target.relposition, distance_upper_bound=maxdis)
            if result[1] >= instrument.fiberset.nfibers:
                _logger.debug('no robot nearest to point within %f', maxdis)
                continue
            val = ids[result[1]]
            _logger.debug('robot is number % d', val)
            robot = instrument.get_device('MEGARA.MOS.RoboticPositioner_%d' % val)
            fibid1, allpos1 =  robot.fibers_in_focal_plane()
            tab1 = instrument.focal_plane.filter_visible_fibers(fibid1, allpos1)
            fibid1 = tab1['fibid']
            pos_x = tab1['x']
            pos_y = tab1['y']
            subfinal = final[fibid1 - 1]
            # FIXME: check angles
            rotang = 90 - robot.pa
            subfinal = add_target_mos(target, subfinal, wl, fibrad, pos_x, pos_y, rotang, oe, telescope, atmosphere)
            final[fibid1 - 1] = subfinal

        # Recompute all visible fibers
        fibid, allpos = fibermos.fibers_in_focal_plane()
        tab = instrument.focal_plane.filter_visible_fibers(fibid, allpos)
        fibid = tab['fibid']
        pos_x = tab['x']
        pos_y = tab['y']
        cover_frac = tab['cover']
        subfinal = final[fibid - 1]
    else:
        _logger.debug('simulate LCB')
        lcb = instrument.get_device('MEGARA.LCB')
        fibid, allpos = lcb.fibers_in_focal_plane()
        tab = instrument.focal_plane.filter_visible_fibers(fibid, allpos)
        fibid = tab['fibid']
        pos_x = tab['x']
        pos_y = tab['y']
        cover_frac = tab['cover']
        subfinal = final[fibid - 1]
        add_targets_lcb2(t1, subfinal, wl, fibrad, pos_x, pos_y, oe, telescope, atmosphere)

    add_sky(t2, subfinal, wl, fibarea, telescope)

    # Uniform lamp objects
    add_lamp(t3, subfinal, wl, fibarea, pos_x, pos_y)

    subfinal *= cover_frac[:, np.newaxis]

    final[fibid-1] = subfinal

    return wl, final


def add_targets_lcb(targets, subfinal, wl, fibrad, pos_x, pos_y, oe, telescope, atmosphere):

    airmass = oe.airmass
    extinction = np.power(10, -0.4 * airmass * atmosphere.extinction(wl))
    seeing_fwhm = atmosphere.seeing.fwhm(wl[0], oe.zenith_distance)
    _logger.debug('seeing FWHM is %', seeing_fwhm)
    # This is not really correct, but is simpler
    # than doing the whole PSF convolution
    _logger.debug('internal focus factor %s', telescope.focus_actuator.internal_focus_factor)
    # Scale with focus
    seeing_fwhm_focus = seeing_fwhm * telescope.focus_actuator.internal_focus_factor
    telescope.focus_actuator.focus = 1200
    print('intnl focus factor', telescope.focus_actuator.internal_focus_factor)
    seeing_profile = atmosphere.seeing.profile(seeing_fwhm_focus)
    psf = None
    fraction_of_flux = simulate_point_like_profile(seeing_profile, psf, fibrad.value)

    energy_unit = u.erg * u.s**-1 * u.cm**-2 * u.AA **-1
    photon_energy = (cons.h * cons.c / wl)

    for target in targets:
        _logger.debug('object name is %s', target.name)
        center = target.relposition
        # fraction of flux in each fiber
        #scales = np.zeros((nfibers,))

        # Offset fiber positions
        offpos0 = pos_x - center[0]
        offpos1 = pos_y - center[1]

        f_o_f = fraction_of_flux(offpos0, offpos1)
        # If spectrum is in erg s^-1 cm^-2 AA^-1
        # Handle units
        sed = target.spectrum['sed'](wl) * extinction * energy_unit

        nphotons = sed / photon_energy * telescope.area * telescope.transmission(wl)
        subfinal += f_o_f[:, np.newaxis] * nphotons.to(u.s**-1 * u.micron**-1)

    return subfinal


def add_targets_lcb2(targets, subfinal, wl, fibrad, pos_x, pos_y, oe, telescope, atmosphere):
    """This correction includes DAR"""
    ref_wl = 0.5 * (wl.max() + wl.min())
    _logger.debug('reference wl is %s', ref_wl)
    seeing_fwhm = atmosphere.seeing.fwhm(ref_wl, oe.zenith_distance)
    _logger.debug('seeing FWHM is %s', seeing_fwhm)
    # Scale with focus
    # telescope.focus_actuator.focus = 3000
    _logger.debug('internal focus factor %s', telescope.focus_actuator.internal_focus_factor)
    seeing_fwhm_focus = seeing_fwhm * telescope.focus_actuator.internal_focus_factor
    #
    seeing_profile = atmosphere.seeing.profile(seeing_fwhm_focus)
    psf = None
    fraction_of_flux = simulate_point_like_profile(seeing_profile, psf, fibrad.value)

    airmass = oe.airmass
    zenith_distance = oe.zenith_distance
    extinction = np.power(10, -0.4 * airmass * atmosphere.extinction(wl))

    dar = atmosphere.refraction(zenith_distance, wl, ref_wl).to(u.arcsec).value
    _logger.debug('extreme DAR %s %s', dar[0], dar[-1])

    energy_unit = u.erg * u.s**-1 * u.cm**-2 * u.AA **-1
    photon_energy = (cons.h * cons.c / wl)

    for target in targets:
        _logger.debug('object is %s', target.name)
        center = target.relposition

        # Offset fiber positions
        # Assume DAR affects only one coordinate, this depends on the several
        # different angles
        center_dar = np.zeros((2, wl.size))
        center_dar[0] = center[0] + 0.0 * dar
        center_dar[1] = center[1] + 1.0 * dar
        offpos1_dar = pos_y[:, None] - center_dar[1]
        offpos0_dar = pos_x[:, None] - center_dar[0]

        offpos0 = pos_x - center[0]
        offpos1 = pos_y - center[1]
        f_o_f_dar = fraction_of_flux(offpos0_dar, offpos1_dar)
        f_o_f_wl = f_o_f_dar.sum(axis=0)
        # print f_o_f_dar
        f_o_f = fraction_of_flux(offpos0, offpos1)
        _logger.debug('fraction of flux recovered (no DAR) %f', f_o_f.sum())
        _logger.debug('fraction of flux recovered min %f', f_o_f_wl.min())
        _logger.debug('fraction of flux recovered max %f', f_o_f_wl.max())

        # If spectrum is in erg s^-1 cm^-2 AA^-1
        # Handle units
        sed = target.spectrum['sed'](wl) * extinction * energy_unit

        nphotons = sed / photon_energy * telescope.area * telescope.transmission(wl)
        flux_dar = f_o_f_dar * nphotons.to(u.s**-1 * u.micron**-1)
        # flux_no_dar = f_o_f[:, np.newaxis] * nphotons.to(u.s ** -1 * u.micron ** -1)
        subfinal += flux_dar

    return subfinal


def add_target_mos(target, subfinal, wl, fibrad, pos_x, pos_y, rotang, oe, telescope, atmosphere):

    ref_wl = 0.5 * (wl.max() + wl.min())
    _logger.debug('reference wl is %s', ref_wl)

    seeing_fwhm = atmosphere.seeing.fwhm(ref_wl, oe.zenith_distance)
    _logger.debug('seeing FWHM is %s', seeing_fwhm)
    # telescope.focus_actuator.focus = 3000
    _logger.debug('internal focus factor %s', telescope.focus_actuator.internal_focus_factor)
    seeing_fwhm_focus = seeing_fwhm * telescope.focus_actuator.internal_focus_factor
    seeing_profile = atmosphere.seeing.profile(seeing_fwhm_focus)
    psf = None
    fraction_of_flux = simulate_point_like_profile(seeing_profile, psf, fibrad.value, angle=rotang, xsize=5.0, ysize=5.0)

    airmass = oe.airmass
    _logger.debug('airmass is %s', airmass)
    zenith_distance = oe.zenith_distance
    _logger.debug('Z dis is %s', zenith_distance)

    extinction = np.power(10, -0.4 * airmass * atmosphere.extinction(wl))

    energy_unit = u.erg * u.s**-1 * u.cm**-2 * u.AA **-1
    photon_energy = (cons.h * cons.c / wl)

    _logger.debug('object is %s', target.name)
    center = target.relposition

    ref_wl = 0.5 * (wl.max()  + wl.min())

    _logger.debug('reference wl is %s', ref_wl)
    dar = atmosphere.refraction(zenith_distance, wl, ref_wl).to(u.arcsec).value
    _logger.debug('extreme DAR %s %s', dar[0], dar[-1])
    # Offset fiber positions
    # Assume DAR affects only one coordinate, this depends on the several
    # different angles
    center_dar = np.zeros((2, wl.size))
    center_dar[0] = center[0] + 0.0 * dar
    center_dar[1] = center[1] + 1.0 * dar
    offpos1_dar = pos_y[:, None] - center_dar[1]
    offpos0_dar = pos_x[:, None] - center_dar[0]

    offpos0 = pos_x - center[0]
    offpos1 = pos_y - center[1]
    f_o_f_dar = fraction_of_flux(offpos0_dar, offpos1_dar)
    f_o_f_wl = f_o_f_dar.sum(axis=0)
    # print f_o_f_dar
    f_o_f = fraction_of_flux(offpos0, offpos1)
    _logger.debug('fraction of flux recovered (no DAR) %f', f_o_f.sum())
    _logger.debug('fraction of flux recovered min %f', f_o_f_wl.min())
    _logger.debug('fraction of flux recovered max %f', f_o_f_wl.max())

    # If spectrum is in erg s^-1 cm^-2 AA^-1
    # Handle units
    sed = target.spectrum['sed'](wl) * extinction * energy_unit

    nphotons = sed / photon_energy * telescope.area * telescope.transmission(wl)
    flux_dar = f_o_f_dar * nphotons.to(u.s**-1 * u.micron**-1)
    # flux_no_dar = f_o_f[:, np.newaxis] * nphotons.to(u.s ** -1 * u.micron ** -1)
    subfinal += flux_dar

    return subfinal


# Uniform atmosphere objects
def add_sky(targets, subfinal, wl, fibarea, telescope):

    energy_unit = u.erg * u.s**-1 * u.cm**-2 * u.AA**-1 * u.arcsec**-2

    photon_energy = (cons.h * cons.c / wl)
    _logger.debug('add sky targets')
    for target in targets:
        sed = target(wl) * energy_unit
        nphotons = sed / photon_energy * fibarea * telescope.area * telescope.transmission(wl)
        subfinal += nphotons.to(u.s**-1 * u.micron**-1)

    return subfinal


# Uniform lamp objects
def add_lamp(lamps, subfinal, wl, fibarea, pos_x, pos_y):

    photon_energy = (cons.h * cons.c / wl)

    for lamp in lamps:
        scales = np.ones_like(pos_x)
        if lamp.illumination:
            scales *= lamp.illumination(pos_x, pos_y)

        sed = lamp.flux(wl)
        area = 1.0 * u.cm ** 2  # FIXME, lamp surface area?
        nphotons = (sed / photon_energy * area * fibarea).to(u.s ** -1 * u.micron ** -1)
        subfinal += scales[:, np.newaxis] * nphotons[np.newaxis, :]

    return subfinal



#
# Copyright 2015 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# Megara DRP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Megara DRP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Megara DRP.  If not, see <http://www.gnu.org/licenses/>.
#

from numina.core.taggers import get_tags_from_full_ob


def tagger_empty(obsres):
    return {}


def tagger_vph(obsres):
    return get_tags_from_full_ob(obsres, reqtags=['vph'])


def tagger_base_image(obsres):
    return get_tags_from_full_ob(obsres, reqtags=['vph', 'insmode'])
